# coding: utf-8
#
#    TypeAtlas Sample Text Utilities
#
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

"""Language sample database for testing fonts."""

# TODO: Order: Prefer known languages, then languages in different
#              writing system, *then* languages with yet unseen characters,
#              then all the rest. get_samples_with_prio() that reorders them.
# TODO: Implement fallbacks
# TODO: Match ISO script with the full name from PropertyValueAliases.txt
# TODO: Select languages from aspell, ispell, hunspell dictionaries.
# TODO: Add writing system samples. Add numbers, related to writing systems.
# TODO: Currency
# TODO: OpenType features when displaying pangrams
# TODO: Display all scripts. Need IPA samples.

from __future__ import unicode_literals

from collections import namedtuple, defaultdict, OrderedDict
from itertools import chain
import sys
from typeatlas.util import OrderedSet, N_, warnmsgf, generic_type
# RFC 3066 code pangrams

SetOf = generic_type('Set')
SequenceOf = generic_type('Sequence')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
Literal = generic_type('Literal')
Optional = generic_type('Optional')
Union = generic_type('Union')

ANY = object()
ALL = object()

LangListType = Union[SequenceOf[str], Literal[ANY, ALL]]
ScriptListType = SequenceOf[str]
ScriptType = Union[str, Literal[ANY, ALL]]


DEFAULT = ''
ALTERNAITVE = '-alt'
DATED = '-dated'
ANCIENT = '-ancient'


PRIO_PREFER = 150
PRIO_DEFAULT = 100
PRIO_WITHHOLD = 50

class SampleType(object):

    """A type of sample with its feature. For example, a pangram
    would contain all vowels, consonants, etc."""

    def __init__(self, name: str,
                       multichar: bool=True, phrase: bool=False,
                       all_characters: bool=False,
                       all_consonants: bool=False,
                       all_vowels: bool=False,
                       all_accents: bool=False,
                       all_accent_combinations: bool=False,
                       both_cases: bool=False,
                       all_digits: bool=False,
                       all_symbols: bool=False,
                       all_syllables: bool=False,
                       all_strokes: bool=False,
                       extensive_coverage: bool=False,
                       incomplete: bool=False,
                       common_only: bool=False,
                       no_repeats: bool=False):
        self.name = name
        self.multichar = multichar
        self.phrase = phrase
        self.all_characters = all_characters
        self.all_consonants = all_consonants
        self.all_vowels = all_vowels
        self.all_accents = all_accents
        self.all_accent_combinations = all_accent_combinations
        self.both_cases = both_cases
        self.all_symbols = all_symbols
        self.all_syllables = all_syllables
        self.all_strokes = all_strokes
        self.extensive_coverage = extensive_coverage
        self.incomplete = incomplete
        self.common_only = common_only
        self.no_repeats = no_repeats

    def variant(self, name, **kwargs) -> 'SampleType':
        """Return a variant of that type."""
        kw = dict(vars(self))
        kw.update(name=name, **kwargs)
        return type(self)(**kw)


CHARACTER = SampleType(N_('Character'), multichar=False)
PHRASE = SampleType(N_('Phrase'), phrase=True)
PANGRAM = PHRASE.variant(N_('Pangram'),
                         all_characters=True,
                         all_consonants=True,
                         all_vowels=True,
                         all_accents=True,
                         all_accent_combinations=True,
                         all_strokes=True)

PERFECT_PANGRAM = PANGRAM.variant(N_("Perfect pangram"),
                                  no_repeats=True)

KOREAN_PANGRAM = PANGRAM.variant(N_("Phrase with all vowels and consonants"),
                                 all_characters=False)

KOREAN_CONSONANT_PANGRAM = KOREAN_PANGRAM.variant(
                                N_("Phrase with all consonants"),
                                all_vowels=False)

# Those are different
PANGRAM_PARTIAL_ACCENTS = PANGRAM.variant(N_("Pangram without all accents"),
                                          all_accents=False,
                                          all_accent_combinations=False)
PANGRAM_LIMITED_ACCENTS = PANGRAM.variant(N_("Pangram without all accent combinations"),
                                          all_accents=False)
SYLLABLE_PANGRAM = PANGRAM.variant(N_('Syllable pangram'),
                                   all_syllables=True,
                                   all_characters=False)
PARTIAL_PANGRAM = PANGRAM.variant(N_('Partial pangram'),
                                  incomplete=True)

YONG_CHARACTER = CHARACTER.variant(N_('Yong character'),
                                   all_strokes=True,
                                   common_only=True)


SampleInfo = namedtuple('SampleInfo',
                        ['code', 'text', 'script', 'origin', 'sources',
                         'english', 'translit', 'original', 'version',
                         'type', 'note', 'flags', 'prio'])

SampleInfo.lang = property(lambda self: self.code)


def normalize_langcode(langcode: str) -> str:
    """Turn the langcode into something the modules of this module like.

    get_sample/get_samples/has_sample call this for you.
    """

    langcode = langcode.lower().replace('_', '-')
    langcode = langcode.partition('@')[0].partition('.')[0]
    return langcode


class LanguageSamples(object):

    """A memory database of loaded language sample texts."""

    def __init__(self):
        #self.codes_with_pangrams = set()
        #self.code_scripts_with_pangrams = set()
        self.samples_regular = OrderedDict()
        self.samples_long = OrderedDict()
        self.display_scripts = OrderedDict()
        self.sample_fallbacks = OrderedDict()
        self.unsorted_samples = []

    normalize_code = staticmethod(normalize_langcode)
    sample_factory = staticmethod(SampleInfo)

    def use_display_scripts(self, code: str, scripts: SequenceOf[str]):
        """Register the given scripts for use with the given language code."""
        if code in self.display_scripts:
            scripts = OrderedSet(chain(self.display_scripts[code], scripts))
        self.display_scripts[code] = tuple(scripts)

    def use_fallback(self, requested: str, fallback_to: str,
                           script: ScriptType=ANY):
        """Register the second language as a fallback for the first, for the
        given script, when looking up samples."""
        if requested in self.sample_fallbacks:
            fbs = self.sample_fallbacks[requested]
        else:
            fbs = ()

        fbs += (fallback_to, )
        self.sample_fallbacks[requested] = fbs

    def has_sample(self, code: str, exact: bool=False, fallbacks: bool=True,
                         script: ScriptType=ANY, long=False) -> bool:
        """Return True if we have sample for lang code.

        Options mean the same as for get_sample.
        """

        samples = self.samples_long if long else self.samples_regular

        if not exact:
            code = self.normalize_code(code)

        if script is ANY:
            if code in samples:
                return True
            if fallbacks and code.partition('-')[0] in samples:
                return True
            return False

        else:
            if fallbacks:
                code_alt = code.partition('-')[0]
                keys = ((code, ANY), (code_alt, ANY),
                        (code, script), (code_alt, script))
            else:
                keys = ((code, ANY), (code, script))

            return any(key in samples for key in keys)


    def get_sample(self, code: str, exact: bool=False, fallbacks: bool=True,
                         alternative: int=0, long: bool=False,
                         script: ScriptType=ANY) -> Optional[SampleInfo]:
        """Get a sample for the given language code.

        If alternative > 0 is passed, an alternative phrase may be returned.
        If long=True is passed, the function is allowed to return long phrases.
        If script is passed, return the phrase for the given script.

        If exact is True, the language code won't be normalized.
        If fallbacks is False, only the exact language will be returned, i.e.
        if a region is specified and there isn't a sample for this region, it
        won't be returned.
        """
        samples = self.samples_long if long else self.samples_regular

        if not exact:
            code = self.normalize_code(code)

        if script is ANY:
            results = samples.get(code)
            if not results and fallbacks:
                results = samples.get(code.partition('-')[0])

        else:
            results = samples.get((code, script))

            # Preferring ANY script over fallback. Reason:
            # ANY script is a fallback for samples where script is undefined.
            # We just assume it matches, so we prefer it. That may be wrong, though.
            if not results:
                results = samples.get((code, ANY))

            if not results and fallbacks:
                results = samples.get((code.partition('-')[0], script))

            if not results and fallbacks:
                results = samples.get((code.partition('-')[0], ANY))

        if alternative is ALL:
            return list(results or ())
        if alternative is ANY:
            alternative = 0
        if results:
            return results[min(alternative, len(results) - 1)]

        return None

    def get_samples(self, codes: LangListType,
                          exact: bool=False, fallbacks: bool=True,
                          alternative: int=0, long=False,
                          script: ScriptType=ALL,
                          scripts: ScriptListType=None) -> IteratorOf[SampleInfo]:
        """Yield a sample for each of the given language codes.

        ANY can be passed to get any, if you don't know the language.

        You can pass script to limit

        If alternative > 0 is passed, alternative phrases may be returned.
        So of multiple samples are supported for the same language, you can
        pass alternative=1, 2, 3, etc. to get an alternative one. This can
        be used to vary the result to e.g. build longer text.

        If long=True is passed, the function is allowed to return long phrases.
        If script is passed, return phrases for the given script. Does not make
        sense to pass this.

        The exact argument controls the interpretation of the language codes,
        preventing normalization. See get_sample() for exact, fallbacks, etc.
        """

        if codes is ANY:
            codes = self.available_language_codes(long=long, script=script)

        for code in codes:
            if scripts is None:
                if script is ALL:
                    scripts = self.display_scripts.get(code) or [ANY]
                else:
                    scripts = [script]

            seen = set()

            for scr in scripts:
                sample = self.get_sample(code,
                                         exact=exact, fallbacks=fallbacks,
                                         alternative=alternative, long=long,
                                         script=scr)

                # If all alternatives are requested, don't care if we
                # already returned that.
                if alternative is ALL:
                    yield from sample
                    continue

                if sample is None:
                    continue

                if sample.code in seen:
                    continue
                if (sample.code, sample.script) in seen:
                    continue

                if sample.script is ANY:
                    seen.add(sample.code)
                else:
                    seen.add((sample.code, sample.script))

                yield sample

    def add_sample(self, code: str, text: str, script: ScriptType=ANY,
                         origin: str=None, sources: SequenceOf[str]=(),
                         english: str=None, translit: str=None, original: str=None,
                         version: str=DEFAULT, type: SampleType=PHRASE, note: str='',
                         flags: IterableOf[str]=frozenset(),
                         prio: int=PRIO_DEFAULT,
                         langpos: int=None, langscriptpos: int=None) -> SampleInfo:

        """Register the given sample for use with the given language.

        You can provide text, script, origin, sources where the sample was taken from,
        enligh version, translit version (in Latin script), original phrase,
        version of the sample, type of the sample (e.g. PANGRAM), any note,
        flags that define its use, and priority (e.g. prio PRIO_WITHHOLD would
        preclude its immediate use).


        You can insert later samples at earlier positions using langpos
        and langscriptpos arguments. That's considered a bad hack.
        """

        if code == 'en' or code.startswith('en'):
            if not english:
                english = text

        code = self.normalize_code(code)

        #code_scripts_with_pangrams.add((code, script))
        #codes_with_pangrams.add(code)

        sampleinfo = self.sample_factory(code, text, script, origin,
                                         tuple(sources),
                                         english, translit, original,
                                         version, type,
                                         note, frozenset(flags), int(prio))

        self.unsorted_samples.append(sampleinfo)

        if 'no-display' in sampleinfo.flags or 'non-default' in sampleinfo.flags:
            return sampleinfo

        if 'long-only' in sampleinfo.flags:
            if script != 'Brai':
                self.samples_long.setdefault(code, []).append(sampleinfo)
            self.samples_long.setdefault((code, script), []).append(sampleinfo)
            return sampleinfo

        # Don't use Braille samples by default
        if script != 'Brai':
            samples = self.samples_regular.setdefault(code, [])
            if langpos is None or langpos >= len(samples):
                samples.append(sampleinfo)
            else:
                samples.insert(langpos, sampleinfo)

        samples = self.samples_regular.setdefault((code, script), [])
        if langscriptpos is None or langscriptpos >= len(samples):
            samples.append(sampleinfo)
        else:
            samples.insert(langscriptpos, sampleinfo)

        return sampleinfo

    def available_language_codes(self, long: bool=None,
                                       script: ScriptType=ALL) -> IteratorOf[str]:
        """Get all the available language codes.

        The script argument is ignored for now, but it is supported
        so it can be passed.
        """
        if long is None:
            samples = self.samples_regular.keys() | self.samples_long.keys()
        else:
            samples = self.samples_long if long else self.samples_regular
        for code in samples:
            if isinstance(code, tuple):
                continue
            yield code


main_samples = LanguageSamples()

add_sample = main_samples.add_sample
has_sample = main_samples.has_sample
get_sample = main_samples.get_sample
get_samples = main_samples.get_samples
use_fallback = main_samples.use_fallback
use_display_scripts = main_samples.use_display_scripts
available_language_codes = main_samples.available_language_codes


def load_samples():
    """Load language samples bundled with typeatlas."""
    from typeatlas.data import samples_language
    from typeatlas.data import samples_language_pango
    from typeatlas.data import samples_language_glass
    from typeatlas.data import samples_braille
    samples_language, samples_language_pango, samples_language_glass
    samples_braille


def load_glass():
    """Load eat glass samples. Currently does nothing."""
    from typeatlas.data.samples_language_glass import EAT_GLASS_ENGLISH, EAT_GLASS_SOURCE

    eat_glass = {}

    for key, value in eat_glass.items():
        if key not in main_samples.samples_regular:
            warnmsgf("No translation for %r, eating glass", key)
            add_sample(key, value,
                       english=EAT_GLASS_ENGLISH,
                       sources=[EAT_GLASS_SOURCE])


def load_nonfree():
    """Load pangram data whose licensing is unclear, and suspected to be
    non-free, but can be legally distributed with TypeAtlas for at least
    non-commercial purposes."""

    return

    try:
        from typeatlas.data.samples_language_nonfree \
            import hovercraft, HOVERCRAFT_ENGLISH, \
                   HOVERCRAFT_TRANSLATION_SOURCE, HOVERCRAFT_NOTE

    except ImportError:
        return

    for key, value in hovercraft.items():
        if key not in main_samples.samples_regular:
            warnmsgf("No translation for %r, using hovercraft", key)
            add_sample(key, value,
                       english=HOVERCRAFT_ENGLISH,
                       sources=[HOVERCRAFT_TRANSLATION_SOURCE],
                       note=HOVERCRAFT_NOTE)


load_samples()
#load_nonfree()
#load_glass()


def _filter_samples():
    """Remove unwanted samples."""
    for values in chain(main_samples.samples_long.values(),
                        main_samples.samples_regular.values()):
        while len(values) > 1 and values[0].prio < PRIO_DEFAULT:
            del values[0]

_filter_samples()



# -*- coding: utf-8 -*-
#
#    TypeAtlas External Commands
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

from collections import namedtuple, OrderedDict
import subprocess as sp
import shutil
import shlex
import os
import os.path
import sys
import traceback

import typeatlas
from collections.abc import Callable
from typeatlas.util import generic_type

Union = generic_type('Union')
Optional = generic_type('Optional')
Literal = generic_type('Literal')
AnyStr = generic_type('AnyStr')
SequenceOf = generic_type('Sequence')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
MappingOf = generic_type('Mapping')
TupleOf = generic_type('Tuple')


STOP = object()

DEFAULT_NEWLINE = b'\n'
N_ = lambda s: s
_UNSPECIFIED = object()

ON_DEMAND = object()

commands_by_provisions = {}
commands_by_executable = {}


def commands_providing(executor: 'Executor',
                       provision: str) -> 'IteratorOf[ExternalCommand]':
    """Yield all external commands providing the requested provision
    (i.e. program) using the provided executor.

    For example, commands_providing(Executor(), 'fc-list') yield
    commands that call fc-list locally using subprocess.Popen().
    And 'ssh-execute' would provide SSH execution, whereas 'font-view'
    would provide commands providing external viewers.
    """

    commands = []
    for cls in commands_by_provisions.get(provision, ()):
        command = cls(executor)
        if command.available():
            yield command


def command_providing(executor: 'Executor',
                      provision: str,
                      default=_UNSPECIFIED) -> 'Optional[ExternalCommand]':
    """Return an external command providing the requested provision
    (i.e. program) using the provided executor.

    For example, command_providing(Executor(), 'fc-list') returns
    a command that call fc-list locally using subprocess.Popen().
    """
    for command in commands_providing(executor, provision):
        return command

    if default is _UNSPECIFIED:
        raise KeyError(provision)
    return default


def register(cls: type) -> type:
    """Register a given class to provide external commands.

    The commands which it provides are taken from its provides attribute,
    which needs to be an iterable.

    This can be used as a decorator."""

    for provision in cls.provides:
        commands_by_provisions.setdefault(provision, []).append(cls)
    commands_by_executable.setdefault(cls.executable, []).append(cls)
    return cls


def get_line_maker_callbacks(line_callback: Callable,
                             exit_callback: Callable,
                             delimiter: bytes=DEFAULT_NEWLINE,
                             ) -> TupleOf[Callable, Callable]:
    """Convert a line and exit callback to data and exit callback.

    You provide a callback that accepts processed lines of bytes as arguments,
    and  a callback that is called on exit or completion of an external command.
    Returned are a callback that accepts chunks of byte data, and anopther callback
    that is called on exit or completion.

    The chunks are merged progressively as the returned data callback is called,
    and any lines found along the way are split and passed to your line callback.

    Your callbacks will be called respectively. Your exit callback will be
    called with the same arguments as the returned one will be, but after
    any remaining lines are passed back to the line callback.

    You can change the delimited from b'\n'.

    This method only supports bytes, supporting strings is left as an exercise
    for the reader.
    """

    buf = [b'']

    def line_wrapper(data: bytes):
        data = buf[0] + data
        lines = data.split(delimiter)
        if data != b'':
            buf[0] = lines.pop()

        for line in lines:
            if line_callback(line) is STOP:
                return STOP

    def exit_wrapper(*args, **kwargs):
        if buf[0]:
            # XXX Should always be one line
            for line in buf[0].split(delimiter):
                line_callback(line)
        return exit_callback(*args, **kwargs)

    return line_wrapper, exit_wrapper


def get_merger_callbacks(data_callback: Callable,
                         exit_callback: Callable,
                         ) -> TupleOf[Callable, Callable]:

    """Convert a callback that receives the entire data, and an exit callback,
    to a pair of callbacks that receives data chunks and handle exit.

    You provide a callback that receives the entire data as bytes or str,
    and a callback called at exit. Returned is a callback that accepts data
    chunks, and one to be called at completion or exit.

    The chunks received are accumulated as the returned data callback is
    called, until completion, when a single merged string is passed to
    your data_callback, and then your exit callback is called with the same
    arguments as the original one was.

    If NO CHUNKS are received, your data callback is not called at all.

    This method supports both str and bytes.
    """

    buf = []

    def data_wrapper(data: AnyStr):
        buf.append(data)

    def exit_wrapper(*args, **kwargs):
        if buf and isinstance(buf, str):
            data_callback(''.join(buf))
        else:
            data_callback(b''.join(buf))
        return exit_callback(*args, **kwargs)

    return data_wrapper, exit_wrapper


class Call(object):

    """A call to an external commands, containing arguments, environment,
    and CallCallbacks instance to process the result."""

    def __init__(self, args: SequenceOf[AnyStr],
                       env: MappingOf[AnyStr, AnyStr]=None,
                       postprocess_callbacks: 'CallCallbacks'=None):
        self.args = args
        self.env = env
        self.postprocess_callbacks = postprocess_callbacks


class CallDefinition(object):

    """A special descriptor to decorate a method of ExternalCommand for
    calling external commands. Use the cls.make() decorator.

    The decorated method will return the arguments or Call() instance, and
    if an executor is present, it will be executed. Otherwise, the arguments
    will be returned.

    A translatable description can be added for each definition, which
    will be used in the GUI for locating external commands. As well as the
    name of an argument (presently unused).

    If wait=True is passed, the process is waited for, and postprocessing
    is performed. If wait=ON_DEMAND is passed to the class, the actual
    waiting is up to the caller.

    For commands expecting result, wait=True is to be used. For commands
    that are started (e.g. an external font viewer), wait=False (default)
    is to be used. For commands executing other commnads (e.g. ssh),
    wait=ON_DEMAND is to be used.

    The line_output argument is presently unused.
    """

    def __init__(self, function: Callable,
                       description: str=None,
                       wait: 'Union[bool, Literal[ON_DEMAND]]'=False,
                       line_output: bool=True, argument: str=None,
                       postprocess_callbacks: 'CallCallbacks'=None):
        self.function = function
        self.description = description
        self.wait = wait
        self.postprocess_callbacks = postprocess_callbacks

    @classmethod
    def make(cls, *args, **kwargs) -> Callable:
        """Return a decorator that turns a method a call definition, making
        it callable with your loaded executor."""
        def decorator(func):
            return cls(func, *args, **kwargs)
        return decorator

    def __call__(self, instance, *args, **kwargs):
        """Calls the method, processing any output and waiting if requested."""
        exec_kwargs = {}

        callbacks = kwargs.pop('callbacks', instance.executor.callbacks)

        if self.wait is ON_DEMAND:
            exec_kwargs['wait'] = kwargs.pop('wait', False)

        call = self.function(instance, *args, **kwargs)

        if not isinstance(call, Call):
            cmdline_args = call
        else:
            cmdline_args = call.args
            postprocess = call.postprocess_callbacks
            if postprocess is None:
                postprocess = self.postprocess_callbacks
            if postprocess is not None:
                callbacks = postprocess.result_wrapper(callbacks)
            exec_kwargs['env'] = call.env 

        exec_kwargs['callbacks'] = callbacks

        if instance.executor:
            return instance.executor.execute(self, instance,
                                             cmdline_args,
                                             **exec_kwargs)

        return cmdline_args

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return CallDefinitionMethod(self, instance)


class CallDefinitionMethod(object):

    """A bound variant of CallDefinition that calls the call definition
    with self as required."""

    def __init__(self, definition, instance):
        self.definition = definition
        self.instance = instance

    def __getattr__(self, attr):
        return getattr(self.definition, attr)

    def __call__(self, *args, **kwargs):
        return self.definition(self.instance, *args, **kwargs)


RETURN_DEFAULT = object()


class ResultCallbacks(object):

    """Result callbacks for the post-processed output of the commands.

    These are used by command *callers* to get the result and/or error
    asynchronously, providing a result and error callback.

    They are different from the postprocess callbacks used internally to
    postproces the result (see below for CallCallbacks).

    If error_callback is RETURN_DEFAULT (the default), the provided
    default is given to the result_callback(). The exception printed
    on stderr, unless callable error_callback is provided.
    """

    def __init__(self, result_callback: Callable=None,
                       error_callback: Callable=RETURN_DEFAULT,
                       default=None):
        if result_callback is not None:
            self.handle_result = result_callback

        self.default_on_error = False
        if error_callback is RETURN_DEFAULT:
            self.default_on_error = True
        elif error_callback is not None:
            self.handle_error = error_callback

        self.default = default

    def handle_result(self, result):
        """Handle the result. This is replaced by the result_callback if
        provided, and you can override in subclass."""
        return result

    def handle_error(self, exc_type, exc_value, exc_traceback):
        """Handle error. The default prints the exception to stderr,
        and if requested, gives the default to the result handler."""
        traceback.print_exception(exc_type, exc_value, exc_traceback)
        if self.default_on_error:
            return self.handle_result(self.default)


class CallCallbacks(object):

    """Callbacks for post-processing the output to get a single result.


    These are used internally by the commands, and different from the
    result callbacks provided by the command callers to get the result
    asynchronously (see above for ResultCallbacks).

    They can contain a stdout and exit callback, receiving the stdout and
    exit code, respectively, line_mode=True (default) makes sure that
    is provided as lines, and chunked=False and line_mode=False makes
    sure it is merged into a single bytestring instead.

    If merge_outputs=True is passed, stderr is passed to stdout.

    A command can require blocking or non-blocking mode, which is an odd
    thing to do.

    You can provide a bufsize other than the default of 256 KB.
    """

    def __init__(self, stdout_callback: Callable=None,
                       exit_callback: Callable=None,
                       line_mode: bool=True,
                       chunked: bool=True,
                       bufsize: int=256 * 1024,
                       merge_outputs: bool=False,
                       require_blocking: bool=False,
                       require_nonblocking: bool=False):

        if stdout_callback is not None:
            self.process_stdout = stdout_callback
        if exit_callback is not None:
            self.handle_exit = exit_callback

        self.line_mode = line_mode
        self.chunked = chunked
        self.bufsize = bufsize

        self.require_blocking = require_blocking
        self.require_nonblocking = require_nonblocking

        self.merge_outputs = merge_outputs

    def result_wrapper(self, callbacks: ResultCallbacks=None) -> 'CallCallbacks':
        """Wrap ResultCallbacks and return a new CallCallbacks which
        will call the caller's ResultCallbacks callback."""
        if callbacks is None:
            return self

        exit_handler = self.handle_exit

        def exit_wrapper(exitcode):
            try:
                result = exit_handler(exitcode)
            except:
                exc_tuple = sys.exc_info()
                return callbacks.handle_error(*exc_tuple)
            else:
                return callbacks.handle_result(result)

        cls = type(self)

        return cls(self.process_stdout, exit_wrapper,
                   line_mode=self.line_mode, chunked=self.chunked,
                   merge_outputs=self.merge_outputs,
                   require_blocking=self.require_blocking,
                   require_nonblocking=self.require_nonblocking)

    def chunked_wrapper(self) -> 'CallCallbacks':
        """Return a CallCallbacks instance that accepts chunked output, and
        cannot provide line_mode themselves.

        This should be called by executors that provide such output, like Qt."""
        if self.chunked and not self.line_mode:
            return self
        cls = type(self)

        if self.line_mode:
            return cls(*get_line_maker_callbacks(
                              self.process_stdout,
                              self.handle_exit),
                       line_mode=False, chunked=True,
                       merge_outputs=self.merge_outputs,
                       require_blocking=self.require_blocking,
                       require_nonblocking=self.require_nonblocking)
        else:
            return cls(*get_merger_callbacks(
                              self.process_stdout,
                              self.handle_exit),
                       line_mode=False, chunked=True,
                       require_blocking=self.require_blocking,
                       require_nonblocking=self.require_nonblocking)

    @classmethod
    def single(cls, callback: Callable,
                    line_mode: bool=True,
                    chunked: bool=False,
                    *args, **kwargs) -> 'CallCallbacks':
        """Create CallCallbacks instance that accepts both exit code and stdout
        in a single callback."""
        stdout = []

        def stdout_callback(data):
            stdout.append(data)

        def exit_callback(exitcode):
            if chunked or line_mode:
                return callback(exitcode, stdout)
            else:
                return callback(exitcode, stdout[0])

        return cls(stdout_callback, exit_callback,
                   line_mode=line_mode, chunked=chunked, *args, **kwargs)

    def process_stdout(self, data: bytes):
        """Process part of the output or the whole output. This is replaced
        with the stdout  callback provided to the constructor."""
        pass

    def handle_exit(self, exitcode: int):
        """Handle the exit code. This is replaced with the exit callback
        provided to the constructor."""
        return exitcode


class ExecutorCallbacksWrapper(object):

    """An executor wrapper that provides given set of processing CallCallbacks
    callbacks to it automatically for every command.

    This is used by the with_callbacks() executor method, which is not seeing
    much use.
    """

    def __init__(self, executor: 'Executor',
                       callbacks: CallCallbacks=None):
        self.executor = executor
        self.callbacks = callbacks

    def __getattr__(self, attr):
        return getattr(self.executor, attr)

    def execute(self, call: CallDefinition,
                      command: 'ExternalCommand',
                      args: SequenceOf[AnyStr],
                      callbacks: CallCallbacks=_UNSPECIFIED,
                      *a, **kw):
        if callbacks is _UNSPECIFIED:
            callbacks = self.callbacks

        return self.executor.execute(call, command, args,
                                     callbacks=callbacks,
                                     *a, **kw)


class ExecutionModeUnsupportedError(RuntimeError):
    pass


class BlockingModeUnsupported(ExecutionModeUnsupportedError):
    pass


class NonblockingModeUnsupported(ExecutionModeUnsupportedError):
    pass


class Executor(object):

    """An executor for commands that uses Python's subprocess by default.

    It accepts custom_paths to specify the full path to a command executable,
    or to forward it to a wrapper found in your operating system.

    It can also include default CallCallbacks which is seldom used as default ones.
    """

    def __init__(self, custom_paths: MappingOf[str, str]={},
                       callbacks: CallCallbacks=None, *args, **kwargs):
        super(Executor, self).__init__(*args, **kwargs)
        self.custom_paths = custom_paths
        self.callbacks = callbacks

    def with_callbacks(self, callbacks) -> ExecutorCallbacksWrapper:
        """Get a wrapped executor with changed default callbacks."""
        return ExecutorCallbacksWrapper(self, callbacks)

    def get_executable_path(self, executable: str) -> Optional[str]:
        """Get the path of a given executable."""
        path = self.custom_paths.get(executable)
        if path is not None and os.access(path, os.X_OK):
            return path
        return shutil.which(executable)

    def executable_available(self, executable_path: Optional[str]) -> bool:
        """Return True if the executable is available (its path is not
        None)."""
        return executable_path is not None

    def execute(self, call: CallDefinition,
                      command: 'ExternalCommand',
                      args: SequenceOf[AnyStr],
                      env: MappingOf[AnyStr, AnyStr]=None,
                      callbacks: CallCallbacks=_UNSPECIFIED,
                      wait: 'Union[bool, Literal[ON_DEMAND]]'=_UNSPECIFIED):
        """Execute the given command, using the provided definition,
        external command instance, arguments, environment, processing
        callbacks (that *can* wrap result callbacks that will call the
        original caller's callbacks asynchronously)."""

        if callbacks is _UNSPECIFIED:
            callbacks = self.callbacks

        if callbacks.require_nonblocking:
            raise NonblockingModeUnsupported('subprocess is always blocking')

        if wait is _UNSPECIFIED:
            wait = call.wait

        if env is not None:
            procenv = dict(os.environ)
            procenv.update(env)
        else:
            procenv = None

        popen_kwargs = {}

        if wait and callbacks is not None:
            popen_kwargs['stdout'] = sp.PIPE
            if callbacks.merge_outputs:
                popen_kwargs['stderr'] = sp.STDOUT

        p = sp.Popen(args, env=procenv, **popen_kwargs)

        if wait:
            if callbacks is not None:
                if callbacks.line_mode:
                    for line in p.stdout:
                        callbacks.process_stdout(line.rstrip(DEFAULT_NEWLINE))
                elif callbacks.chunked:
                    while True:
                        buf = p.stdout.read(callbacks.bufsize)
                        if not buf:
                            break
                        callbacks.process_stdout(buf)
                else:
                    callbacks.process_stdout(p.stdout.read())

            exitcode = p.wait()

            if callbacks is not None:
                return callbacks.handle_exit(exitcode)
            return exitcode


class ExternalCommand(object):

    """An external command, like ssh, fc-list

    The command calls are provided by the class, and it is instantiated
    with an executor and custom executable path. The latter is filled
    automatically from the executor if not provided.
    """

    kind = 'command'
    provides = []

    executable = None
    icon = None

    def __init__(self, executor: Executor, executable_path: str=None):
        if not executable_path:
            executable_path = executor.get_executable_path(self.executable)
        self.executable_path = executable_path
        self.executor = executor

    def available(self) -> bool:
        """Return True if the command is available with the given executor."""
        return self.executor.executable_available(self.executable_path)


class SshExecutor(Executor):

    """An executor that uses SSH to execute commands remotely on a host."""

    def __init__(self, ssh: 'SshCommand', host: str,
                       callbacks: CallCallbacks=None):
        self.ssh = ssh
        self.host = host
        self.callbacks = callbacks

    def get_executable_path(self, executable: str) -> Optional[str]:
        # Opportunistic approach - hope it's in PATH on remote side
        return executable

    def executable_available(self, executable_path: str) -> bool:
        # Opportunistic approach - assume its available if ssh is available
        return self.ssh.available()

    def execute(self, call: CallDefinition,
                      command: ExternalCommand,
                      args: SequenceOf[AnyStr],
                      callbacks: CallCallbacks=_UNSPECIFIED,
                      wait: 'Union[bool, Literal[ON_DEMAND]]'=_UNSPECIFIED):

        if callbacks is _UNSPECIFIED:
            callbacks = self.callbacks

        if wait is not _UNSPECIFIED:
            wait = call.wait

        return self.ssh.remote_execute(self.host, args,
                                       wait=wait,
                                       callbacks=callbacks)


@register
class SshCommand(ExternalCommand):

    """SSH command for executing other commands remotely. That's used
    by the SshExecutor."""

    provides = ['ssh-executor', 'ssh-execute',
                'remote-executor', 'remote-execute']
    executable = 'ssh'

    def get_remote_executor(self, host):
        return SshExecutor(self, host)

    get_ssh_executor = get_remote_executor

    @CallDefinition.make(description=N_("Execute command with OpenSSH"),
                         wait=ON_DEMAND)
    def remote_execute(self, host: str, args: SequenceOf[str]):
        """Execute a remote command"""
        args = [arg if isinstance(arg, str) else arg.decode('ascii')
                for arg in args]

        return [self.executable_path, '-o', 'BatchMode=yes', '--', host,
                ' '.join(shlex.quote(arg) for arg in args)]

    ssh_execute = remote_execute

PackageFormat = namedtuple('PackageFormat', 'ext description icon')
PackageResult = namedtuple('PackageResult', 'format name')


package_formats = {
    'deb': PackageFormat('deb', N_('Debian package'), 'application-x-deb'),
    'rpm': PackageFormat('deb', N_('RPM package'), 'application-x-rpm'),
}


def _dpkg_callbacks(filename: str=None) -> CallCallbacks:

    """Return callbacks for parsing dpkg output."""

    result = OrderedDict()

    def line_callback(line):
        package, sep, filename = line.decode('utf8').partition(':')
        if not sep: 
            return
        filename = filename.lstrip().rstrip('\n')
        result[filename] = PackageResult('deb', package.strip())

    def exit_callback(exitcode):
        if filename is not None:
            result.get(filename)
        return result

    return CallCallbacks(line_callback, exit_callback, 
                         line_mode=True)


@register
class DpkgCommand(ExternalCommand):

    """A dpkg command for discovering package information."""

    provides = ['find-file-package', 'find-files-packages', 'dpkg']
    executable = 'dpkg'

    @CallDefinition.make(description=N_("Find the package containing file"),
                         wait=True)
    def find_file_package(self, filename: str):
        """Find the package containing given file."""
        return Call([self.executable_path, '-S', '--', filename],
                    postprocess_callbacks=_dpkg_callbacks(filename),
                    env=dict(LC_ALL='C.UTF-8'))

    @CallDefinition.make(description=N_("Find the package containing file"),
                         wait=True)
    def find_files_packages(self, filenames: IterableOf[str]):
        """Find the packages containing a given iterable of files."""
        return Call([self.executable_path, '-S', '--'] + list(filenames),
                    postprocess_callbacks=_dpkg_callbacks(),
                    env=dict(LC_ALL='C.UTF-8'))


def _rpm_callbacks(filenames: IterableOf[str],
                   filename: str=None) -> CallCallbacks:

    """Return callbacks for parsing rpm output."""

    # FIXME: WRONG?

    result = OrderedDict()
    filenames = iter(filenames)

    def line_callback(line):
        try:
            filename = next(filenames)
        except StopIteration:
            return
        if line.startswith(b'error: ') or b'not owned' in line:
            return
        package = line.decode('utf8').strip()
        result[filename] = PackageResult('rpm', package)

    def exit_callback(exitcode):
        if filename is not None:
            result.get(filename)
        return result

    return CallCallbacks(line_callback, exit_callback, 
                         line_mode=True, merge_outputs=True)


@register
class RpmCommand(ExternalCommand):

    """A rpm command for discovering package information."""

    provides = ['find-file-package', 'find-files-packages', 'rpm']
    executable = 'rpm'

    @CallDefinition.make(description=N_("Find the package containing file"),
                         wait=True)
    def find_file_package(self, filename: str):
        """Find the package containing given file."""
        return Call([self.executable_path, '-q', '-f', '--', filename],
                    postprocess_callbacks=_rpm_callbacks([filename], filename),
                    env=dict(LC_ALL='C.UTF-8'))

    @CallDefinition.make(description=N_("Find the package containing file"),
                         wait=True)
    def find_files_packages(self, filenames: SequenceOf[str]):
        """Find the packages containing a given iterable of files."""
        return Call([self.executable_path, '-q', '-f', '--'] + list(filenames),
                    postprocess_callbacks=_rpm_callbacks(filenames),
                    env=dict(LC_ALL='C.UTF-8'))


@register
class FcListCommand(ExternalCommand):

    """A fc-list command for querying the complete fontconfig font list."""

    provides = ['fc-list']
    executable = 'fc-list'

    @CallDefinition.make(wait=True)
    def custom(self, args: IterableOf[AnyStr]):
        """Run fc-list with custom arguments."""
        return [self.executable_path] + list(args)


@register
class FcMatchCommand(ExternalCommand):

    """A fc-list command for performing custom matches against the fontconfig
    font list."""

    provides = ['fc-match']
    executable = 'fc-match'
    
    @CallDefinition.make(wait=True)
    def custom(self, args: IterableOf[AnyStr]):
        """Run fc-list with custom arguments."""
        return [self.executable_path] + list(args)


@register
class FontForgeCommand(ExternalCommand):

    """Access to any installed fontforge for font opening and editing."""

    provides = ['font-edit', 'font-open']
    executable = 'fontforge'
    icon = 'fontforge'

    @CallDefinition.make(description=N_("Edit with FontForge"), 
                         argument='font')
    def edit(self, font: 'typeatlas.fontlist.FontLike'):
        """Edit font with font forge."""
        return [self.executable_path, os.path.abspath(font.file)]

    open = edit


@register
class KFontViewCommand(ExternalCommand):

    """Access to any installed kfontview for font opening and viewing."""

    provides = ['font-view', 'font-open']
    executable = 'kfontview'
    icon = 'kfontview'

    @CallDefinition.make(description=N_("View with KFontView"),
                         argument='font')
    def view(self, font: 'typeatlas.fontlist.FontLike'):
        """View font with kfontview."""
        return [self.executable_path, os.path.abspath(font.file)]

    open = view


@register
class GnomeFontViewerCommand(ExternalCommand):

    """Access to any installed gnome-font-viewer for font opening
    and viewing."""

    provides = ['font-view', 'font-open']
    executable = 'gnome-font-viewer'
    icon = 'fonts'

    @CallDefinition.make(description=N_("View with GNOME Font Viewer"),
                         argument='font')
    def view(self, font: 'typeatlas.fontlist.FontLike'):
        """View font with gnome-font-view."""
        return [self.executable_path, os.path.abspath(font.file)]

    open = view


@register
class GnomeCharMapCommand(ExternalCommand):

    """Access to any installed gucharmap for font opening and viewing."""

    provides = ['font-view', 'font-open']
    executable = 'gucharmap'
    icon = 'gucharmap'

    @CallDefinition.make(description=N_("Open GNOME Character Map with "
                                        "this font"),
                         argument='font')
    def view(self, font: 'typeatlas.fontlist.FontLike'):
        """View font with gucharmap."""
        return [self.executable_path, '--font', font.fullname + ' 24']

    open = view


@register
class FontyPythonCommand(ExternalCommand):

    """Access to any installed fontypython for font directory viewing."""

    provides = ['directory-open']
    executable = 'fontypython'
    icon = 'fontypython'

    @CallDefinition.make(description=N_("View fonts with Fonty Python"),
                         argument='directory')
    def open(self, path: str):
        """View the directory with fontypython."""
        return [self.executable_path, os.path.abspath(path)]




# -*- coding: utf-8 -*-
#
#    TypeAtlas Generic Utilities
#    Copyright (C) 2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

"""A library that allows one to perform complex operations on discrete ranges
of e.g. codepoints defined by set-like objects that store lists of
first and last number in a range. This is a complex library that
provides full Set interface. If you want simple operations, consider
using blockmath.

Like blocks from blockmath, the ranges are defined as their start and end,
inclusively, unlike Python's range().
"""

# FIXME Too many hardcoded types, particularly RangeWrapper.

from collections.abc import Set, Iterable, Callable
from itertools import chain
from typeatlas import blockmath
import re
import sys
import abc
import bisect
import numbers

from typeatlas.util import generic_type

SequenceOf = generic_type('Sequence')
IterableOf = generic_type('Iterable')
AnyStr = generic_type('AnyStr')
Optional = generic_type('Optional')


range_pattern = re.compile(r'(?s)^(.+?)(?:[:-](.+))?$')


class RangeBase(Set):

    """The base class for discrete ranges. Use this class as a type hint.

    The ranges property defines any subranges.
    """

    __slots__ = ()

    value_type = None
    range_type = None

    ranges = ()

    def to_plain_blocks(self) -> SequenceOf[blockmath.BlockLike]:
        """Convert to blocks (plain tuples)"""
        return [(rg.start, rg.end) for rg in self.ranges]

    def to_blocks(self) -> SequenceOf[blockmath.Block]:
        """Convert to blocks (Block named tuples)"""
        return [blockmath.Block(rg.start, rg.end) for rg in self.ranges]

    def to_fontconfig(self) -> str:
        """Convert to fontconfig format."""
        return ' '.join('%x-%x' % (rg.start, rg.end) for rg in self.ranges)

    def tostring(self) -> str:
        """Convert to decimal string. This is broken for character ranges
        with spaces at the ends."""
        return ' '.join('%s-%s' % (rg.start, rg.end) for rg in self.ranges)

    @abc.abstractmethod
    def successor(self, value: object) -> object:
        """Get the value after a given one, e.g. value + 1"""
        raise NotImplementedError

    @abc.abstractmethod
    def predecessor(self, value: object) -> object:
        """Get the value before a given one, e.g. value - 1"""
        raise NotImplementedError

    @abc.abstractmethod
    def sortkey(self, value: object) -> object:
        """Get the sort value for a given value"""
        raise NotImplementedError


class RangeWrapper(RangeBase):

    """A wrapper of a range. This can be used to lazily parse a range to
    avoid an expensive parsing operation until the range is really needed.
    The real real range is in the wrapped property.
    """

    __slots__ = ('wrapped', )

    @property
    def range_type(self):
        return self.wrapped.range_type

    @property
    def value_type(self):
        return self.wrapped.value_type

    @property
    def ranges(self):
        return self.wrapped.ranges

    def __init__(self, wrapped: RangeBase):
        self.wrapped = wrapped

    def __contains__(self, value):
        return value in self.wrapped

    def __iter__(self):
        return iter(self.wrapped)

    def __reversed__(self):
        return reversed(self.wrapped)

    def __len__(self):
        return len(self.wrapped)

    def successor(self, value: object) -> object:
        return self.wrapped.successor(value)

    def predecessor(self, value: object) -> object:
        return self.wrapped.predecessor(value)

    def sortkey(self, value: object) -> object:
        return self.wrapped.sortkey(value)

    def from_iterable(self, it: Iterable) -> RangeBase:
        return self.wrapped.from_iterable(it)

    def _from_iterable(self, it: Iterable) -> Set:
        return self.wrapped._from_iterable(it)

    def to_ordinal_range(self) -> 'RangeBase':
        return self.wrapped.to_ordinal_range()

    def to_character_range(self) -> 'RangeBase':
        return self.wrapped.to_character_range()

    def __eq__(self, other):
        return self.wrapped == other

    def __ne__(self, other):
        return self.wrapped != other

    def __le__(self, other):
        return self.wrapped <= other

    def __ge__(self, other):
        return self.wrapped >= other

    def isdisjoint(self, other):
        return self.wrapped.isdisjoint(other)

    def __or__(self, other):
        return self.wrapped | other

    def __ror__(self, other):
        return other | self.wrapped

    def __and__(self, other):
        return other & self.wrapped

    def __rand__(self, other):
        return self.wrapped & other

    def __sub__(self, other):
        return self.wrapped - other

    def __rsub__(self, other):
        return other - self.wrapped

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.wrapped)

    def __getattr__(self, attr):
        return getattr(self.wrapped, attr)


class LazyRangeWrapper(RangeWrapper):

    """A lazy range wrapper that uses a factory to reinitialize
    the range, for example for belated parsing."""

    __slots__ = ('factory', 'args', 'kwargs')

    def __init__(self, factory: Callable, *args, **kwargs):
        self.factory = factory
        self.args = args
        self.kwargs = kwargs

    def __getattr__(self, attr):
        if attr == 'wrapped':
            self.wrapped = self.factory(*self.args, **self.kwargs)
            del self.factory, self.args, self.kwargs
            if self.wrapped is None:
                # non-strict mode
                self.wrapped = EmptyRange()
            return self.wrapped

        return super().__getattr__(attr)

        #raise AttributeError('%r object has no attribute %r'
        #                        % (type(self).__name__, attr))


class EmptyRange(RangeBase):

    """An empty range. This needs to be explicit, because of the way
    ranges are defined it is impossible to define the range as empty."""

    __slots__ = ()

    def __contains__(self, value):
        return False

    def __iter__(self):
        return iter(())

    def __reversed__(self):
        return iter(())

    def __len__(self):
        return 0

    def successor(self, value: object) -> object:
        raise TypeError("empty ranges don't have values")

    def predecessor(self, value: object) -> object:
        raise TypeError("empty ranges don't have values")

    def sortkey(self, value: object) -> object:
        raise TypeError("empty ranges don't have values")

    @classmethod
    def _from_iterable(cls, it: Iterable) -> Set:
        return frozenset(it)

    def __eq__(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, (Range, MultiRange)):
            return False
        if isinstance(other, EmptyRange):
            return True
        return super().__le__(other)

    def __le__(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, (Range, MultiRange)):
            return True
        return super().__le__(other)

    def __ge__(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, (Range, MultiRange)):
            return False
        return super().__ge__(other)

    def __repr__(self):
        return 'EMPTY_RANGE'

    def to_ordinal_range(self) -> 'RangeBase':
        return self

    def to_character_range(self) -> 'RangeBase':
        return self


EMPTY_RANGE = EmptyRange()


class Range(RangeBase):

    """A basic discrete range with start and end. The start and end are
    inclusive, unlike Python's range()."""

    __slots__ = ('start', 'end')

    def __init__(self, start: object, end: object):
        self.start = start
        self.end = end

        if start not in self:
            raise ValueError("not %r <= %r" % (start, end))

    @property
    def ranges(self) -> 'SequenceOf[Range]':
        """A sequence of self"""
        return (self, )

    @property
    def range_type(self):
        return type(self)

    def __repr__(self):
        return "%s(%r, %r)" % (type(self).__name__, self.start, self.end)

    def __eq__(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, Range):
            return self.start == other.start and self.end == other.end
        return super().__eq__(other)

    def __le__(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, Range):
            return self.start in other and self.end in other
        return super().__le__(other)

    def __ge__(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, Range):
            return other.start in self and other.end in self
        return super().__ge__(other)

    def isdisjoint(self, other):
        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, Range):
            if self.start in other:
                return False
            if self.end in other:
                return False
            if other.start in self:
                return False
            if other.end in self:
                return False
            return True

        return super().isdisjoint(other)

    @classmethod
    def from_blocks(cls, it: IterableOf[tuple]) -> RangeBase:
        ranges = [cls(block[0], block[1]) for block in it]
        if not ranges:
            return EMPTY_RANGE
        elif len(ranges) == 1:
            return ranges[0]
        else:
            return MultiRange(ranges)

    @classmethod
    def from_iterable(cls, it: Iterable) -> RangeBase:
        """Create the range from an iterable. If non-contiguous,
        a MultiRange may be returned."""
        return NotImplemented

    @classmethod
    def _from_iterable(cls, it: Iterable) -> Set:
        """Create a set from an iterable. This either returns a range
        or a frozenset."""
        result = cls.from_iterable(it)
        if result is NotImplemented:
            return frozenset(it)
        return result

    @classmethod
    def fromstring(cls, s: AnyStr, lazy: bool=True,
                                   strict: bool=True) -> Optional[RangeBase]:
        """Create the range from the string returned by tostring().
        If non-contiguous, a MultiRange may be returned."""
        if lazy:
            if not strict and s is None:
                return None
            if not s:
                return EMPTY_RANGE
            return LazyRangeWrapper(cls.fromstring, s,
                                    lazy=False, strict=strict)
        if not strict:
            if s is None:
                return None
            try:
                cls.fromstring(s, lazy=False, strict=True)
            except ValueError:
                return None

        if not isinstance(s, str):
            s = s.decode('utf8')

        ranges = []
        for part in s.split():
            match = range_pattern.search(part)
            if match is None:
                raise ValueError(repr(part))

            start, end = match.groups()

            start = cls.value_type(start)
            if end is None:
                ranges.append(cls(start, start))
            else:
                end = cls.value_type(end)
                ranges.append(cls(start, end))

        if not ranges:
            return EMPTY_RANGE
        if len(ranges) == 1:
            return ranges[0]
        return MultiRange(ranges)

    def __and__(self, other):

        if isinstance(other, Range):
            start = end = None
            if self.start in other:
                start = self.start
                end = other.end
            elif other.start in self:
                start = other.start
                end = self.end

            if self.end in other:
                end = self.end
                if start is None:
                    start = other.start
            elif other.end in self:
                end = other.end
                if start is None:
                    start = self.start

            if start is None or end is None:
                return EMPTY_RANGE

            return type(self)(start, end)

        if isinstance(other, EmptyRange):
            return other

        if isinstance(other, MultiRange):
            return NotImplemented

        return super().__and__(other)

    __rand__ = __and__

    def __or__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            start = end = None
            if self.start in other:
                start = other.start
                end = self.end
            elif other.start in self:
                start = self.start
                end = other.end

            if self.end in other:
                end = other.end
                if start is None:
                    start = self.start
            elif other.end in self:
                end = self.end
                if start is None:
                    start = other.start

            if start is None or end is None:
                return MultiRange([self, other])

            return type(self)(start, end)

        if isinstance(other, EmptyRange):
            return self

        if isinstance(other, MultiRange):
            return NotImplemented

        return super().__or__(other)

    __ror__ = __or__

    def __sub__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            start = end = None
            if self.start in other:
                if other.end in self:
                    start = other.successor(other.end)
                    if start not in self:
                        return EMPTY_RANGE
                    return type(self)(start, self.end)

                else:
                    return EMPTY_RANGE

                start = other.start
                end = self.end

            elif self.end in other:
                if other.start in self:
                    end = other.predecessor(other.start)
                    if end not in self:
                        return EMPTY_RANGE
                    return type(self)(self.start, end)

            elif other.start in self:
                assert other.end in self

                p1 = other.predecessor(other.start)
                p2 = other.successor(other.end)

                assert p1 in self
                assert p2 in self

                return MultiRange([type(self)(self.start, p1),
                                   type(self)(p2, self.end)])

            else:
                assert other.end not in self
                return self

        if isinstance(other, EmptyRange):
            return self

        if isinstance(other, MultiRange):
            return NotImplemented

        return super().__sub__(other)

    def __rsub__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            return type(self).__sub__(other, self)

        if isinstance(other, MultiRange):
            return NotImplemented

        if isinstance(other, EmptyRange):
            return self

        return super().__rsub__(other)


class OrdinalRange(Range):

    """A range of numbers, where the successor and predecessor are computed
    with value + 1 and value - 1."""

    __slots__ = ()

    value_type = int

    def __contains__(self, value):
        return self.start <= value <= self.end

    def __iter__(self):
        return iter(range(self.start, self.end + 1))

    def __reversed__(self):
        return iter(range(self.end, self.start - 1, -1))

    def __len__(self):
        return self.end - self.start + 1

    def successor(self, value: numbers.Real) -> numbers.Real:
        return value + 1

    def predecessor(self, value: numbers.Real) -> numbers.Real:
        return value - 1

    def sortkey(self, value: numbers.Real) -> numbers.Real:
        return value

    @classmethod
    def from_fontconfig(cls, s: AnyStr, lazy: bool=True,
                             strict: bool=False) -> Optional[RangeBase]:
        """Create the range from fontconfig character range data.
        If non-contiguous (likely), a MultiRange may be returned."""
        if lazy:
            if not strict and s is None:
                return None
            if not s:
                return EMPTY_RANGE
            return LazyRangeWrapper(cls.from_fontconfig, s,
                                    lazy=False, strict=strict)
        if not strict:
            if s is None:
                return None

            try:
                cls.from_fontconfig(s, lazy=False, strict=True)
            except ValueError:
                return None

        if not isinstance(s, bytes):
            s = s.encode('ascii')

        ranges = []
        for part in s.split():
            start, sep, end = part.partition(b'-')
            start = int(start, 16)
            if not sep:
                ranges.append(cls(start, start))
            else:
                end = int(end, 16)
                ranges.append(cls(start, end))

        if not ranges:
            return EMPTY_RANGE
        if len(ranges) == 1:
            return ranges[0]
        return MultiRange(ranges)

    @classmethod
    def from_iterable(cls, it: IterableOf[numbers.Real]) -> RangeBase:
        """Create the range from an iterable. If non-contiguous,
        a MultiRange may be returned."""
        previous = rg = None
        ranges = []
        for value in sorted(it):
            if previous != value - 1:
                rg = cls(value, value)
                ranges.append(rg)
            elif value == previous + 1:
                rg.end = value
            previous = value

        if len(ranges) == 1:
            return ranges[0]
        elif not ranges:
            return EMPTY_RANGE
        else:
            return MultiRange(ranges)

    def to_ordinal_range(self) -> 'RangeBase':
        """Get the equivalent ordinal range."""
        return self

    def to_character_range(self) -> 'RangeBase':
        """Get the equivalent character range."""
        return CharacterRange(chr(self.start), chr(self.end))


class CharacterRange(Range):

    """A range of unicode characters, where the successor and predecessor are
    computed with codepoint + 1 and codepoint - 1."""

    __slots__ = ("start_ord", "end_ord")

    value_type = str

    def __init__(self, start: str, end: str, *args, **kwargs):
        self.start_ord = ord(start)
        self.end_ord = ord(end)
        super().__init__(start, end, *args, **kwargs)

    def __contains__(self, value):
        return self.start_ord <= ord(value) <= self.end_ord

    def __iter__(self):
        return map(chr, range(self.start_ord, self.end_ord + 1))

    def __reversed__(self):
        return map(chr, range(self.end_ord, self.start_ord - 1, -1))

    def __len__(self):
        return self.end_ord - self.start_ord + 1

    def successor(self, value: str) -> str:
        return chr(ord(value) + 1)

    def predecessor(self, value: str) -> str:
        return chr(ord(value) - 1)

    def sortkey(self, value: str) -> int:
        return ord(value)

    @classmethod
    def from_iterable(cls, it: IterableOf[str]) -> RangeBase:
        """Create the range from an iterable. If non-contiguous,
        a MultiRange may be returned."""
        previous = rg = None
        ranges = []
        for value in sorted(it):
            value_ord = ord(value)

            if previous != value_ord - 1:
                rg = cls(value, value)
                ranges.append(rg)
            elif value_ord == previous + 1:
                rg.end = value
                rg.end_ord = value_ord

            previous = value_ord

        if len(ranges) == 1:
            return ranges[0]
        elif not ranges:
            return EMPTY_RANGE
        else:
            return MultiRange(ranges)

    def to_ordinal_range(self) -> 'RangeBase':
        """Get the equivalent ordinal range."""
        return OrdinalRange(self.start_ord, self.end_ord)

    def to_character_range(self) -> 'RangeBase':
        """Get the equivalent character range."""
        return self


class MultiRange(RangeBase):

    """A non-contiguous discrete set of multiple discrete ranges. This supports
    the same operation as the Range() as is returned whenever an operation between
    them would produce a non-contiguous range."""

    __slots__ = ('ranges', 'starts', 'length')

    def __init__(self, ranges: IterableOf[RangeBase]=()):
        ranges = sorted(filter(None, ranges),
                        key=lambda rg: rg.sortkey(rg.start))
        new_ranges = []
        for rg in ranges:
            if new_ranges and rg.start in new_ranges[-1]:
                new_ranges[-1] |= new_ranges[-1] | rg
            else:
                new_ranges.append(rg)

        self.ranges = tuple(new_ranges)
        self.starts = tuple(rg.sortkey(rg.start) for rg in new_ranges)
        self.length = sum(len(rg) for rg in new_ranges)

    @property
    def range_type(self):
        return type(self.ranges[0])

    @property
    def value_type(self):
        return self.range_type.value_type

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.ranges)

    def __contains__(self, value):
        ranges = self.ranges
        starts = self.starts

        if not ranges:
            return False

        value_key = ranges[0].sortkey(value)

        i = bisect.bisect(starts, value_key) - 1
        if 0 <= i < len(ranges):
            return value in ranges[i]

    def __iter__(self):
        for rg in self.ranges:
            yield from rg

    def __reversed__(self):
        for rg in reversed(self.ranges):
            yield from reversed(rg)

    def __len__(self):
        return self.length

    def successor(self, value: object) -> object:
        return self.ranges[0].successor(value)

    def predecessor(self, value: object) -> object:
        return self.ranges[0].predecessor(value)

    def sortkey(self, value: object) -> object:
        return self.ranges[0].sortkey(value)

    def _from_iterable(self, it: Iterable) -> Set:
        if not self.ranges:
            return frozenset(it)
        result = self.ranges[0].from_iterable(it)
        if result is NotImplemented:
            return frozenset(it)
        return result

    def __eq__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, MultiRange):
            return self.ranges == other.ranges
        if isinstance(other, Range):
            return self.ranges == [other]
        return super().__eq__(other)

    def __le__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped
        if isinstance(other, Range):
            #return bool(self - other)
            return self <= MultiRange([other])

        if isinstance(other, MultiRange):
            pile = list(reversed(self.ranges))
            antipile = list(reversed(other.ranges))

            while pile and antipile:

                first = pile[-1]
                antifirst = antipile[-1]

                if (first.sortkey(first.start) <
                    antifirst.sortkey(antifirst.start)):

                    return False

                elif (antifirst.sortkey(antifirst.end) <
                      first.sortkey(first.start)):

                    antipile.pop()

                elif first.isdisjoint(antifirst):


                    return False

                else:

                    if (first.sortkey(first.end) <=
                        antifirst.sortkey(antifirst.end)):

                        pile.pop()

                    else:
                        modified = first - antifirst
                        if not isinstance(modified, MultiRange):
                            pile[-1] = modified
                        else:
                            # This needs to return False
                            pile[-1:] = reversed(modified.ranges)

                        antipile.pop()

            return not pile

        return super().__le__(other)

    def __ge__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            return bool(other - self)

        if isinstance(other, MultiRange):
            return type(self).__le__(other, self)

        return super().__ge__(other)

    def isdisjoint(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            first = bisect.bisect(self.starts, other.sortkey(other.start)) - 1
            end = bisect.bisect(self.starts, other.sortkey(other.end)) - 1
            return all(rg.isdisjoint(other) for rg in self.ranges[first:end + 1])

        if isinstance(other, MultiRange):
            pile1 = list(reversed(self.ranges))
            pile2 = list(reversed(other.ranges))

            while pile1 and pile2:

                first1 = pile1[-1]
                first2 = pile2[-1]

                if first1.isdisjoint(first2):
                    if first1.sortkey(first1.end) < first2.sortkey(first2.start):
                        pile1.pop()
                    else:
                        assert (first2.sortkey(first2.end) <
                                first1.sortkey(first1.start))
                        pile2.pop()
                else:
                    return False

            return True

        return super().isdisjoint(other)

    def __or__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, MultiRange):
            return type(self)(chain(self.ranges, other.ranges))

        if isinstance(other, Range):
            return type(self)(chain(self.ranges, [other]))

        if isinstance(other, EmptyRange):
            return self

        return super().__or__(other)

    __ror__ = __or__

    def __and__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            first = bisect.bisect(self.starts, other.sortkey(other.start)) - 1
            end = bisect.bisect(self.starts, other.sortkey(other.end)) - 1
            return type(self)((rg & other) for rg in self.ranges[first:end + 1])

        if isinstance(other, MultiRange):
            pile1 = list(reversed(self.ranges))
            pile2 = list(reversed(other.ranges))

            new_ranges = []

            while pile1 and pile2:

                first1 = pile1[-1]
                first2 = pile2[-1]

                if first1.isdisjoint(first2):
                    if (first1.sortkey(first1.end) <
                        first2.sortkey(first2.start)):
                        pile1.pop()
                    else:
                        assert (first2.sortkey(first2.end) <
                                first1.sortkey(first1.start))
                        pile2.pop()
                else:
                    new_ranges.append(first1 & first2)
                    if (first1.sortkey(first1.end) <
                        first2.sortkey(first2.end)):
                        pile1.pop()
                    else:
                        pile2.pop()

            if not new_ranges:
                return EMPTY_RANGE
            if len(new_ranges) == 1:
                return new_ranges[0]
            return type(self)(new_ranges)

        if isinstance(other, EmptyRange):
            return other

        return super().__and__(other)

    __rand__ = __and__

    def __sub__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, Range):
            ##first = bisect.bisect(self.starts, other.sortkey(other.start)) - 1
            ###end = bisect.bisect(self.starts, other.sortkey(other.end)) - 1
            # TODO: Rewrite
            return type(self)(chain.from_iterable((rg - other).ranges
                                                  for rg in self.ranges))

        if isinstance(other, MultiRange):
            pile = list(reversed(self.ranges))
            antipile = list(reversed(other.ranges))

            new_ranges = []

            while pile and antipile:

                first = pile[-1]
                antifirst = antipile[-1]

                if first.isdisjoint(antifirst):
                    if (first.sortkey(first.end) <
                        antifirst.sortkey(antifirst.start)):

                        new_ranges.append(first)
                        pile.pop()

                    else:
                        assert (antifirst.sortkey(antifirst.end) <
                                first.sortkey(first.start))
                        antipile.pop()
                else:
                    if (first.sortkey(first.end) <=
                        antifirst.sortkey(antifirst.end)):

                        new_ranges.append(first - antifirst)
                        pile.pop()

                    else:
                        modified = first - antifirst
                        if not isinstance(modified, MultiRange):
                            pile[-1] = modified
                        else:
                            pile[-1:] = reversed(modified.ranges)
                        antipile.pop()

            if pile:
                new_ranges.extend(pile)

            return type(self)(new_ranges)

        if isinstance(other, EmptyRange):
            return self

        return super().__sub__(other)

    def __rsub__(self, other):

        if isinstance(other, RangeWrapper):
            other = other.wrapped

        if isinstance(other, MultiRange):
            return type(self).__sub__(other, self)

        if isinstance(other, Range):
            return type(self).__sub__(MultiRange([other]), self)

        if isinstance(other, EmptyRange):
            return self

        return super().__rsub__(other)

    def to_ordinal_range(self) -> 'RangeBase':
        """Get the equivalent ordinal range."""
        return MultiRange(rg.to_ordinal_range() for rg in self.ranges)

    def to_character_range(self) -> 'RangeBase':
        """Get the equivalent character range."""
        return MultiRange(rg.to_character_range() for rg in self.ranges)



# -*- coding: utf-8 -*-
#
#    TypeAtlas CLI Charset Functions
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas command-line)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import division

import sys
import argparse
import unicodedata

from typeatlas.langutil import _, N_, textlang, LanguageDatabase
from typeatlas.charinfo import CharacterDatabase, all_character_codes
from typeatlas.charsets import get_encoding


def get_encoding_failing(name, *args, **kwargs):
    result = get_encoding(name, *args, **kwargs)
    if result is None:
        raise KeyError(name)
    return result


charset_argparser = argparse.ArgumentParser(_('TypeAtlas character map'))

charset_argparser.add_argument(
        '--script', '-s',
        help=_('Display the given script'))

charset_argparser.add_argument(
        '--charset', '-C', type=get_encoding_failing,
        help=_('Display the given character set'))

charset_argparser.add_argument(
        '--block', '-B',
        help='Unicode block')

charset_argparser.add_argument(
        'needle', nargs='*',
        help=_('Search for characters related to a word'))


def charwidth(char):
    """Return the width of the character."""
    if unicodedata.east_asian_width(char) in 'WF':
        return 2
    category = unicodedata.category(char)
    if category == 'Cc':
        return -1
    if char in '\x00\u034F\u200B\u200F\u2028\u2029\u202A\u202E\u2060\u2063':
        return 0
    if unicodedata.combining(char):
        return 0
    return 1


class CharPrinter(object):

    def __init__(self, output=sys.stdout):
        self.output = output
        self._chardb = None
        self._langdb = None

    @property
    def langdb(self):
        if self._langdb is None:
            self._langdb = LanguageDatabase.get_instance(populated=True)
        return self._langdb

    @property
    def chardb(self):
        if self._chardb is None:
            self._chardb = CharacterDatabase.get_instance(populated=True)
        return self._chardb

    def print_table_header(self):
        print('%12s' % ('', ), end='', file=self.output)

        for i in range(0, 0xf + 1):
            print('  %1X' % (i, ), end='', file=self.output)

        print(file=self.output)

    def print_table(self, characters):

        row = -float('inf')

        for i, char in characters:
            if not (row <= i <= row + 0xf):
                newrow = i & ~0xf

                if abs(row - newrow) > 0x10:
                    if row > -float('inf'):
                        print(file=self.output)
                        print(file=self.output)
                    self.print_table_header()
                else:
                    print(file=self.output)

                row = newrow

                print('%11X_' % (row >> 4, ), end='', file=self.output)

                for j in range(row, i):
                    print(' %*s' % (2, ''), end='', file=self.output)

            #info = self.chardb.get(ord(char))
            #if info.display:
            #    char = chr(info.display)

            width = charwidth(char)
            if unicodedata.category(char) == 'Cc':
                print(' %*s' % (2, ''), end='', file=self.output)
            elif unicodedata.combining(char):
                print(' %1s%s' % ('', 'o' + char), end='', file=self.output)
            else:
                print(' %*s%s' % (2 - width, '', char), end='',
                      file=self.output)

        print(file=self.output)

    def print_character(self, code, char):
        if self.chardb is None:
            self.chardb = CharacterDatabase.getinstance(populated=True)

        info = self.chardb.get(ord(char), autofill=True)
        script = self.chardb.find_script_name(self.langdb, ord(char))
        plane = self.chardb.get_plane(ord(char))

        print ("%15s: %X" % (_("Hex code"), code), file=self.output)
        print ("%15s: %s" % (_("Character"), char), file=self.output)
        print ("%15s: %s" % (_("Category"), info.category_name(translate=_)),
            file=self.output)
        print ("%15s: %s" % (_("Block"), info.block.name), file=self.output)
        print ("%15s: %s" % (_("Script"), script), file=self.output)
        print ("%15s: %d - %s" % (_("Plane"), plane.number,
                                  _(plane.description)),
               file=self.output)

        for alias in info.aliases:
            print('= ' + alias, file=self.output)
        for alias in info.formalaliases:
            print(info.formal_alias + ' ' + alias, file=self.output)
        for variation in info.variations:
            print ("%15s: %s" % (_("Variation"), variation), file=self.output)

    def select_characters(self, args):
        codes = {}
        selectors = []

        if args.script:
            script = self.chardb.get_alias_value('sc', args.script)
            selectors.append(lambda c: self.chardb.find_script(c) == script)
        if args.block:
            block = args.block
            selectors.append(lambda c: self.chardb.get(c).block == block)

        if not selectors and not args.charset:
            return ((c, chr(c)) for c in range(128))

        elif args.charset is not None:
            charset = args.charset.complete()
            characters = map(ord, charset.characters)
        else:
            charset = None
            characters = all_character_codes()

        for selector in selectors:
            characters = filter(selector, characters)

        if charset is not None:
            return sorted((charset.get_integer(c), chr(c)) for c in characters)

        return sorted((c, chr(c)) for c in characters)


def display_characters(*a, **kw):

    """Command line utility to display character maps."""

    args = charset_argparser.parse_args(*a, **kw)
    printer = CharPrinter()
    characters = printer.select_characters(args)

    printer.print_table(printer.select_characters(args))


if __name__ == '__main__':
    display_characters()



# -*- coding: utf-8 -*-
#
#    TypeAtlas Event / Observer Pattern Signals
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#


"""Signal implementation that is similar (but different) from Qt (PySide/PyQt)
signals.

The module also provides NoisySequence, NoisyMapping, and OrderedNoisyMapping.

The signals here are not generally compatible with Qt signals, but in
a limited subset of their use they can be used interchangeably, provided:

1. Only methods are used (a Slot decorator is provided, for compatibility, not needed).
2. Types known to Qt are passed to Signal() only (e.g. other Signal features are ignored)
3. Only positional arguments are passed to signals (that's recommended anyway)
4. No bare functions are used as callbacks (unless the behaviour is changed to default
   to hardref=True for non-methods).
5. No Qt signals are connected to these signals directly.

   Qt supports callbacks with less arguments that the signal. We currently do
   support such use case too, but only for Python methods, and that may incur
   minor performance penalty, which can be taxing if there are a lot of signals call.
   You may consider not relying on it, as it makes optimizing the module easily
   (remove f.__code__.co_argcount code, etc.)

It is preferable to use Qt signals in the Qt part of the code, and this signals
in the non-Qt part of the code, except for performance reasons.
"""

# FIXME: Cannot connect with .emit of Qt signals unless they are
# connected with hardref=True

import os
import sys
import types
import weakref
import traceback
import functools
from collections import defaultdict, OrderedDict
from collections.abc import MutableMapping, MutableSequence
from collections.abc import Mapping, Callable, Hashable
from inspect import CO_VARARGS, CO_VARKEYWORDS

try:
    from typing import Optional
except ImportError:
    Optional = None


SIGNAL_TRACE = bool(os.environ.get('TYPEATLAS_DEBUG_SIGNAL_TRACE'))
SIGNAL_DEBUG = SIGNAL_TRACE


_empty_dict = {}


def _unpack_method(func: Callable) -> tuple:

    """Extract method as a tuple. If this is a method, this
    returns a tuple of self, the function implementing the method,
    if it is a regular function, it returns a tuple of the function
    and None."""

    try:
        return func.__self__, func.__func__
    except AttributeError:
        return func, None


class ObWeakRef(weakref.ref):

    """A weak reference to an object that is hashable, using object
    identity as hash and equality. This allows to store metadata for
    any objects with __weakref__ that is automatically cleared when
    the object is collected."""

    __slots__ = "_oid",

    def __new__(type, ob, callback=None):
        self = weakref.ref.__new__(type, ob, callback)
        self._oid = id(ob)

        return self

    def __eq__(self, other):

        if isinstance(other, ObWeakRef):

            if self is other:
                return True

            selfob = self()
            otherob = other()

            if selfob is not otherob:
                return False

            if selfob is None:
                return self._oid == other._oid

            return True

        return NotImplemented

    def __ne__(self, other):
        value = self.__eq__(other)

        if value is NotImplemented:
            return NotImplemented

        return not value

    def __hash__(self):
        return hash(self._oid)


class ObRef(weakref.ref):

    """A hard reference to an object that is hashable, using object
    identity as hash and equality. This can be used the reference
    created by ObWeakRef to be hard, and the object to never be
    collected."""

    __slots__ = ("ob", "_oid")

    def __init__(self, ob, callback=None):
        self.ob = ob
        self._oid = id(ob)

    def __eq__(self, other):

        if isinstance(other, ObRef):
            return self.ob is other.ob

        if isinstance(other, ObWeakRef):
            return self.ob is other() is not None

        return NotImplemented

    def __ne__(self, other):
        value = self.__eq__(other)

        if value is NotImplemented:
            return NotImplemented

        return not value

    def __hash__(self):
        return hash(self._oid)


class SignalMethod(object):

    """A signal bound to an instance of a class."""

    def __init__(self, signal: 'Signal', instance: object):
        self.signal = signal
        self.instance = instance

    def connect(self, *args, **kwargs):
        """Connect this signal from this instance of the class to a
        callable (usually a method, called a Slot in Qt vernacular).

        See Signal.connect() for signature; the instance is passed
        automatically."""

        self.signal.connect(self.instance, *args, **kwargs)

    def connect_wrapped(self, *args, **kwargs):
        """Connect a callable wrapping a method.

        See Signal.connect_wrapped() for what that means and for
        a signature."""

        self.signal.connect_wrapped(self.instance, *args, **kwargs)

    def disconnect(self, *args, **kwargs):
        """Disconnect a callable from the instance's signal."""
        self.signal.disconnect(self.instance, *args, **kwargs)

    def __call__(*args, **kwargs):
        """Emit the signal with the given arguments. This can be
        called as both signal.emit() for compatibility with Qt,
        and with signal() to allow connecting signals to signals, or
        to use them as drop-in replacements of methods.

        All callbacks are called. Any exceptions are printed on stderr.

        The callbacks need to have the same signature as the signal,
        unless pass_self=True is passed, which provides as an extra arguments.
        Similarly to Qt, a Python function with less arguments will receive
        less than arguments than the signal. Using this would prevent
        optimizing this module by stripping that code out (it has
        not been timed).
        """

        self = args[0]
        args = args[1:]

        if SIGNAL_TRACE:
            owner = type(self.instance)
            name = self.signal.__name__
            if not name or name == '<signal>':
                for attr in dir(owner):
                    if getattr(owner, attr, None) is self.signal:
                        self.signal.__name__ = name = attr
                        break

            print('[DEBUG] signal: %s.%s(%s) @ %r' % (
                        owner.__name__, self.signal.__name__,
                        ', '.join([repr(arg) for arg in args] +
                                  ['%s=%r' % (key, arg)
                                   for key, arg in kwargs.items()]),
                        self.instance),
                  file=sys.stderr)

        return self.signal(self.instance, *args, **kwargs)

    emit = __call__

    @property
    def __func__(self):
        return self.signal

    @property
    def __self__(self):
        return self.instance
    

class Signal(object):

    """A signal descriptor that can be added to a class so that the class can emit
    signals of that type.

    For partial compatibility with Qt, you can pass accepted types as positional
    arguments (they are also used in the documentation for the signal). The types
    can be strings for forward references.

    When partial compatibility with Qt is not required, you can pass doc,
    name and signature_func arguments, which are presently used only for
    documentation purposes. If inherit_func_name=True is passed, that
    means that the signal is exactly like the signal (we inherit name, doc).

    You can use alternative construction Signal.from_function to decorate
    a function and have the arguments, types and documentation introspected
    from it.
    """

    SIGNAL_ASSIGNMENTS = [attr for attr in functools.WRAPPER_ASSIGNMENTS
                          if attr not in ['__name__', '__doc__',
                                          '__qualname__']]

    SIGNAL_UPDATES = [attr for attr in functools.WRAPPER_UPDATES
                           if attr not in ['__dict__']]

    def __init__(self, *typelist, name: str=None,
                                  doc: str=None,
                                  signature_func: types.FunctionType=None,
                                  inherit_func_name: bool=False):

        # An ordered dict allows us to order the items, and to
        # add some extra metadata for the items that is not part of the key
        self.callbacks = defaultdict(OrderedDict)
        self.types = types

        self.__name__ = name or '<signal>'
        self.__doc__ = doc or ''

        if signature_func is not None:
            if inherit_func_name:
                functools.update_wrapper(self, signature_func)
            else:
                self.__qualname__ = (signature_func.__name__.rpartition('.')[0]
                                     + '.' + self.__name__)
                functools.update_wrapper(self, signature_func,
                                         self.SIGNAL_ASSIGNMENTS,
                                         self.SIGNAL_UPDATES)

        #arg_tuple = '*args'
        arg_tuple = ', '.join('<' + getattr(t, __name__, '') or repr(t)  + '>'
                              for t in typelist)

        self.__doc__ += ('\n\n' if self.__doc__ else '') + (
            'You can connect to this signal with self.{0}.connect(ob.method), \n'
            'and it can be emitted with self.{0}.emit({1}) or \n'
            'self.{0}({1}). Usually the class is responsible \n'
            'for emitting the signal, unless noted otherwise.'
                            .format(self.__name__, arg_tuple))

    def connect(self, instance: object, callback: Callable,
                      pass_self: bool=False, hardref: bool=False,
                      receiver: object=None,
                      pass_receiver: bool=True, pass_class: bool=False):

        """Connect a method, callback or, in Qt vernacular, slot to this
        signal for a given instance. The bound signals pass the instance
        automatically, as self is for a method.

        You should pass a callable, which is preferred to be a builtin
        bound method of an instantiated object. In that case, a weak reference
        to the object is kept, and the signal is automatically disconnected
        when the instance is deleted. Other callables are also weakly referenced,
        which may not work (the default behaviour may be altered in the future).
        The provided callable needs to have the same signature as the signal,
        although for Python function arguments not accepted by the callable
        can be omitted for compatibility with Qt signals.

        If you pass hardref=True, the callback uses a hard reference, and is not
        automatically disconnected.

        If you pass pass_self=True, the instance is also passed to the callback.
        In Qt, one would use self.sender() to get the sender of the signal,
        since this isn't supported by typeatlas signals, passing the instance
        explicitly is the only way to get the sender.

        You can manually pass a receiver of a signal, a function (unbound method).
        That way, the function will be called as the receiver as self (first argument),
        and the signal disconnected when the receiver is collected.

        Then, you can disabling the passing of the receiver by pass_receiver=False,
        or pass the class of the receiver with pass_class=True.
        """

        if receiver is not None:
            ob = receiver
            if pass_class:
                method = classmethod(callback)
            elif pass_receiver:
                method = callback
            else:
                method = staticmethod(callback)
        else:
            ob, method = _unpack_method(callback)


        # Support missing arguments in receiver methods, like Qt
        # Only for Python functions, inspect.signature() is too slow.
        argnames = argcount = None

        if method is None:
            try:
                argnames = ob.__code__.co_varnames
                argcount = ob.__code__.co_argcount
                flags = ob.__code__.co_flags

            except AttributeError:
                pass

            else:
                if flags & CO_VARARGS:
                    argcount = None
                else:
                    if pass_self:
                        argcount = max(argcount - 1, 0)

                if flags & CO_VARKEYWORDS:
                    argnames = None

        else:
            try:
                argnames = method.__code__.co_varnames
                argcount = method.__code__.co_argcount
                flags = method.__code__.co_flags

            except AttributeError:
                pass

            else:
                if flags & CO_VARARGS:
                    argcount = None
                else:
                    argcount -= 1
                    if pass_self:
                        argcount -= 1
                    argcount = max(argcount, 0)

                if flags & CO_VARKEYWORDS:
                    argnames = None
                else:
                    argnames = argnames[1:]

        def remove(ignored=None):
            cbs = self.callbacks[instanceref]
            cbs.pop((obref, method, pass_self), None)
            if not cbs:
                del self.callbacks[instanceref]

        def remove_all(ignored=None):
            self.callbacks.pop(instanceref, None)

        if hardref:
            obref = ObRef(ob)
        else:
            obref = ObWeakRef(ob, remove)

        instanceref = ObWeakRef(instance, remove_all)

        self.callbacks[instanceref][obref, method,
                                    pass_self] = (argcount, argnames)

    def disconnect(self, instance: object, callback: Callable,
                         pass_self: bool=False,
                         receiver: object=None,
                         pass_receiver: bool=True, pass_class: bool=False):

        """Disconnect a signal. You need to pass the same arguments as for
        connect, expect hardref."""

        if receiver is not None:
            ob = receiver
            if pass_class:
                method = classmethod(callback)
            elif pass_receiver:
                method = callback
            else:
                method = staticmethod(callback)

        else:
            ob, method = _unpack_method(callback)

        obref = ObWeakRef(ob)
        instanceref = ObWeakRef(instance)

        cbs = self.callbacks[instanceref]
        del cbs[obref, method, pass_self]
        if not cbs:
            del self.callbacks[instanceref]

    def __call__(*args, **kwargs):
        """Emit the signal with the given arguments. This can be
        called as both signal.emit() for compatibility with Qt,
        and with signal() to allow connecting signals to signals, or
        to use them as drop-in replacements of methods.

        All callbacks are called. Any exceptions are printed on stderr.

        Calling from the bound signal instance, you only provide the arguments.
        If you need to call Signal().emit(), you need to provider the instance
        of the sender of the signal as first argument.

        The callbacks need to have the same signature as the signal,
        unless pass_self=True is passed, which provides as an extra arguments.
        Similarly to Qt, a Python function with less arguments will receive
        less than arguments than the signal. Using this would prevent
        optimizing this module by stripping that code out (it has
        not been timed).
        """
        self, instance = args[0:2]
        args = args[2:]

        instanceref = ObWeakRef(instance)

        callback_tuples = list(self.callbacks.get(instanceref,
                                                  _empty_dict).items())

        for ((obref, method, pass_self),
             (argcount, argnames)) in callback_tuples:

            ob = obref()

            if ob is None:
                continue

            if method is None:
                callback = ob
            else:
                callback = method.__get__(ob, type(ob))

            passargs = args
            passkwargs = kwargs

            # Support functions with less arguments, like Qt.
            if argcount is not None:
                passargs = passargs[:argcount]

            # Rare usage?
            if argnames is not None and passkwargs:
                passkwargs = {arg: passkwargs[arg]
                              for arg in argnames if arg in passkwargs}

            try:
                if pass_self:
                    callback(instance, *passargs, **passkwargs)
                else:
                    callback(*passargs, **passkwargs)
            except:
                traceback.print_exc()

    emit = __call__

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return SignalMethod(self, instance)

    def connect_wrapped(self, instance: object,
                              callback: Callable,
                              wrapper: Callable=None,
                              *args, **kwargs) -> 'Optional[Callable]':
        """If you want to wrap a callback that you will call once your
        wrapper has been called, use this function, which will either
        return a decorator connecting your wrapper (if not provided),
        or connect it directly.

        The callback is provided to your wrapper as first argument.

        This cannot be disconnected manually, only automatically.
        """

        def decorator(wrapper):
            ob, method = _unpack_method(callback)

            def helper(*args, **kwargs):
                if method is None:
                    original_callback = ob
                else:
                    original_callback = method.__get__(ob, type(ob))

                return wrapper(original_callback, *args, **kwargs)

            self.connect(instance, helper, receiver=ob, *args, **kwargs)

        if wrapper is not None:
            return decorator(wrapper)
        return decorator

    @classmethod
    def from_function(cls, *typelist, **kwargs) -> Callable:
        def decorator(func: types.FunctionType) -> 'Signal':
            return cls(*typelist, signature_func=func, inherit_func_name=True,
                       **kwargs)
        return decorator


def Slot(*args) -> Callable:
    """For drop-in compatibility with Qt's Slot/Signal,
    return a do-nothing decorator.."""

    def decorator(func: Callable) -> Callable:
        return func

    return decorator


class NoisyMapping(MutableMapping):

    """A noisy mapping implementation that emits signals when modified.

    You can also proxy an tentatively immutable dictionary with the
    NoisyMapping.proxy() alternative constructor.

    The factory attribute defines what will be used to implement the
    dictionary that will be wrapped here. The temporary_factory
    is used as a temporary dictionary factory when e.g. replacing
    the proxied dictionary.

    Signals are defined at the bottom.
    """

    factory = dict
    temporary_factory = OrderedDict

    is_noisy_collection = True

    _unfreeze = None

    def __init__(self, *args, **kwargs):
        self._values = self.factory(*args, **kwargs)

    @classmethod
    def proxy(cls, values: Mapping=None, unfreeze: Callable=None):
        """Create a proxy for some dictionary. If unfreeze is passed,
        dictionary will be converted to a mutable one using it on the first
        modification."""
        self = cls()
        self.replace_proxied(values, unfreeze)
        return self

    def replace_proxied(self, values: Mapping=None, unfreeze: Callable=None):
        """Change the dictionary being proxied. If unfreeze is passed,
        dictionary will be converted to a mutable one using it on the first
        modification."""
        if values is None:
            values = {}

        # Use a temporary dictionary to emit signals for the replacement
        self._values = self.temporary_factory(self._values)
        self.clear()
        self.update(values)

        self._values = values
        self._unfreeze = unfreeze

    def __getitem__(self, key):
        return self._values[key]

    def __delitem__(self, key):
        if self._unfreeze is not None:
            self._values = self._unfreeze(self._values)
            self._unfreeze = None

        value = self._values[key]

        self.delete_pending(key, value)
        del self._values[key]
        self.deleted(key, value)

    def __setitem__(self, key: Hashable, value: object):
        if self._unfreeze is not None:
            self._values = self._unfreeze(self._values)
            self._unfreeze = None

        existing = key in self._values
        if existing:
            old_value = self._values[key]
            self.value_change_pending(key, value, old_value)
        else:
            self.add_pending(key, value)

        self._values[key] = value

        if existing:
            self.value_changed(key, value, old_value)
        else:
            self.added(key, value)

    def __iter__(self):
        return iter(self._values)

    def __len__(self):
        return len(self._values)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, dict(self._values))

    delete_pending = Signal(Hashable, object,
                            name='delete_pending', signature_func=__setitem__,
                            doc="A key with a given value is about to be deleted")
    add_pending = Signal(Hashable, object,
                         name='add_pending', signature_func=__setitem__,
                         doc="A key with a given value is about to be added")

    deleted = Signal(Hashable, object,
                     name='deleted', signature_func=__setitem__,
                     doc="A key with a given value was just deleted")
    added = Signal(Hashable, object,
                   name='added', signature_func=__setitem__,
                   doc="A key with a given value was just added")

    def _value_changed_signature(self, key: Hashable,
                                       value: object, old_value: object):
        pass

    value_change_pending = Signal(Hashable, object, object,
                                  name='value_change_pending',
                                  signature_func=_value_changed_signature,
                                  doc="A key is about to be changed to a "
                                      "given value from another value")
    value_changed = Signal(Hashable, object, object,
                           name='value_changed',
                           signature_func=_value_changed_signature,
                           doc="A key was changed to a given from another "
                               "value")

    del _value_changed_signature


class OrderedNoisyMapping(NoisyMapping):

    """An ordered version of a noisy mapping"""

    factory = OrderedDict

    def __eq__(self, other):
        return self._values == other

    def __reversed__(self):
        return reversed(self._values)

    def move_to_end(self, key, last=True):
        """Move the key to end. This does not emit signal, because
        no value is added or removed."""
        self._values.move_to_end(key, last=last)

    def popitem(self, last=True):

        try:
            key = next(reversed(self)) if last else next(iter(self))
        except StopIteration:
            raise KeyError('popitem(): dictionary is empty')

        value = self[key]
        del self[key]

        return key, value


class NoisySequence(MutableSequence):

    """A noisy sequence implementation that emits signals when modified.

    You can also proxy an tentatively immutable list with the
    NoisySequence.proxy() alternative constructor.

    The factory attribute defines what will be used to implement the
    list that will be wrapped here. The temporary_factory
    is used as a temporary list y factory when e.g. replacing
    the proxied sequency.

    Signals are defined at the bottom.
    """

    factory = list
    temporary_factory = list

    is_noisy_collection = True

    _unfreeze = None

    def __init__(self, *args, **kwargs):
        self._values = self.factory(*args, **kwargs)

    @classmethod
    def proxy(cls, values=None, unfreeze=None):
        """Create a proxy for some sequence. If unfreeze is passed,
        sequence will be converted to a mutable one using it on the first
        modification."""
        self = cls()
        self.replace_proxied(values, unfreeze)
        return self

    def replace_proxied(self, values=None, unfreeze=None):
        """Change the sequence being proxied. If unfreeze is passed,
        sequence will be converted to a mutable one using it on the first
        modification."""
        if values is None:
            values = {}

        # Use a temporary sequence to emit signals for the replacement
        self._values = self.temporary_factory(self._values)
        del self[:]
        self[:] = values

        self._values = values
        self._unfreeze = unfreeze

    def __delitem__(self, index):
        if self._unfreeze is not None:
            self._values = self._unfreeze(self._values)
            self._unfreeze = None

        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            indices = range(start, stop, step)
            if step > 0:
                indices = reversed(indices)
            for i in indices:
                del self[i]
            return

        value = self._values[index]
        self.remove_pending(index, value)
        del self._values[index]
        self.removed(index, value)

    def __getitem__(self, index):
        return self._values[index]

    def __len__(self):
        return len(self._values)

    def __setitem__(self, index: int, value: object):
        if self._unfreeze is not None:
            self._values = self._unfreeze(self._values)
            self._unfreeze = None

        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))

            if step != 1:
                value = list(value)
                if len(value) != len(range(start, stop, step)):
                    raise ValueError('attempt to assign sequence of size {} '
                                     'to extended slice of size {}'.format(
                                            len(value),
                                            len(range(start, stop, step))))

            iterator = iter(value)

            for i in range(start, stop, step):
                self[i] = next(iterator)

            for i, v in enumerate(iterator, stop):
                assert step != 1, "impossible code reached"
                self.insert(i, v)
            return

        old_value = self._values[index]
        self.value_change_pending(index, value, old_value)
        self._values[index] = value
        self.value_changed(index, value, old_value)

    def insert(self, index, value):
        if self._unfreeze is not None:
            self._values = self._unfreeze(self._values)
            self._unfreeze = None

        self.insert_pending(index, value)
        self._values.insert(index, value)
        self.inserted(index, value)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, list(self._values))

    def _item_sig(self, index: int, value: object):
        pass

    remove_pending = Signal(int, object,
                            name='remove_pending', signature_func=_item_sig,
                            doc="An index with a given value is about to be "
                                "removed")
    insert_pending = Signal(int, object,
                            name='insert_pending', signature_func=_item_sig,
                            doc="At a given index a given value is about to be "
                                "inserted")

    removed = Signal(int, object,
                     name='removed', signature_func=_item_sig,
                     doc="An index with a given value was removed")
    inserted = Signal(int, object,
                      name='inserted', signature_func=_item_sig,
                      doc="At a given index a given value was inserted")

    def _value_changed_signature(self, index: int,
                                       value: object, old_value: object):
        pass

    value_change_pending = Signal(int, object, object,
                                  name='value_change_pending',
                                  signature_func=_value_changed_signature,
                                  doc="At a given index the value is about to "
                                      "be changed to the given value from the "
                                      "given value")
    value_changed = Signal(int, object, object,
                           name='value_changed',
                           signature_func=_value_changed_signature,
                           doc="At a given index the value was changed "
                               "to the given value from the given "
                               "value")

    del _item_sig, _value_changed_signature


def isnoisy(collection: object) -> bool:
    """Return true if this is a noisy collection."""
    return bool(getattr(collection, 'is_noisy_collection', False))



# -*- coding: utf-8 -*-
#
#    TypeAtlas Font Annotations
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

"""Annotations carry information about terms, objects and entities that can
be displayed to the user."""

from typeatlas.util import N_, generic_type
from collections.abc import Iterable, Callable

IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
TupleOf = generic_type('Tuple')


# TODO: Annotate color fonts. Details about use, e.g. PCF/BDF.
# Annotate extensions and mime types, unicode support, OS/browser support.
# Annotate purpose, annotate non-standard, annotate comments (X11 font?)
# native platform, etc.
# Annotate Adobe SVG viewer in CEF_FILE


PROPER_NAME = 1 << 0
COMMON_NAME = 1 << 1

PREFERRED_NAME = 1 << 2
OFFICIAL_NAME = 1 << 3

STANDARD_NAME_BIT = 1 << 4
STANDARD_NAME = OFFICIAL_NAME | STANDARD_NAME_BIT

PROPRIETARY_NAME_BIT = 1 << 5
PROPRIETARY_NAME = OFFICIAL_NAME | PROPRIETARY_NAME_BIT

TRADEMARK_BIT = 1 << 6
TRADEMARK = PROPER_NAME | PROPRIETARY_NAME | TRADEMARK_BIT

REGISTERED_BIT = 1 << 7
REGISTERED = TRADEMARK | REGISTERED_BIT

UNOFFICIAL_NAME = 1 << 8
GENERIC_NAME = 1 << 9
ALT_NAME = 1 << 10

FULL_NAME = 1 << 11
LEGAL_NAME_BIT = 1 << 12
LEGAL_NAME = FULL_NAME | LEGAL_NAME_BIT
ABBREVIATION = 1 << 13


FORMAT_GENERIC_CATEGORY = 1 << 0
FORMAT_SPECIFIC_CATEGORY = 1 << 1
FORMAT_INSTANCE = 1 << 2

FILE_FORMAT = 1 << 3
DATA_FORMAT = 1 << 4


SPEC_STANDARD = 0
SPEC_RECOMMENDATION = 1 
SPEC_DRAFT = 3

BINDING_DEFAULT = object()
BINDING_LIKELY = object()


display_names = [
    (N_('Alternative name'), ALT_NAME),
    (N_('Standard name'), STANDARD_NAME),
    (N_('Name'), TRADEMARK),
    (N_('Name'), PROPRIETARY_NAME),
    (N_('Name'), OFFICIAL_NAME),
    (N_('Unofficial name'), UNOFFICIAL_NAME),
    (N_('Name'), PROPER_NAME),
    (N_('Name'), COMMON_NAME),
]


class Name(object):

    """Describe one of the names of an entity.

    The flags describe what type of name is that, e.g. TRADERMARK would identify
    this as a trademark (which can be checked with flags & TRADEMARK_BIT), and
    ABBREVIATION would be an abbreviation.

    The entities are a list of Annotation objects describing entities (organizations)
    that pertain to the name, e.g. they are the trademark owner.

    The type_label can be a translatable string that displays a human-readable
    representation of the type of name, especially if it can't be represented
    with the flags. E.g. N_('Metonym').

    The trademarked argument refers to which subset of the name is trademarked.
    """

    def __init__(self, text: str, flags: int=0,
                       entities: 'IterableOf[EntityAnnotation]'=None,
                       type_label: str=None,
                       trademarked: str=None):
        self.text = text
        self.flags = flags
        self.entities = list(entities or ())
        self.type_label = type_label
        self.trademarked = trademarked

    def decorated_text(self) -> str:
        """Return a version of the name with e.g. a trademark symbol"""
        if not (self.flags & TRADEMARK_BIT):
            return self.text
        sym = '®' if (self.flags & REGISTERED_BIT) else '™'
        if self.trademarked and self.trademarked in self.text:
            return self.text.replace(self.trademarked, 
                                     self.trademarked + sym, 1)
        return self.text + sym

    def __str__(self):
        return self.decorated_text()

    def __repr__(self):
        return "<%s %r at %x>" % (type(self).__name__,
                                  self.text, id(self))


def short(name) -> object:
    """A key function that sorts short names before long names,
    used e.g. for the prefer argument of get_name()"""
    return name.flags & ABBREVIATION, not (name.flags & FULL_NAME)


def long(name) -> object:
    """A key function that sorts long names before short names,
    used e.g. for the prefer argument of get_name()"""
    return name.flags & FULL_NAME, not (name.flags & ABBREVIATION)


class Annotation(object):

    """Annotation of a given term, object, or entity. This is the base class.

    It can have multiple names, an icon, flags describing the category of the
    objects, such as FILE_FORMAT to describe file formats, FORMAT_GENERIC_CATEGORY
    to describe a category of formats, and FORMAT_INSTANCE to describe
    the exact instance of a format. Some are filled by subclasses.

    It can have generic entities (organizations) pertaining to the subject
    of the annotation, and more specific entities such as developers,
    standard_entities (e.g. standard organization that approved this
    format as a standard), as well as standard_reference (ID),
    standard_level (SPEC_DRAFT, SPEC_RECOMMENDATION or SPEC_STANDARD),
    and an applicable illustration.

    The parents argument provides a list of annotations describing what
    the annotated thing is a subset of. For example, Type 1 font
    format is a type of an online font.

    The extends argument provides an annotation describing what
    the annotated thing is an extended version of. For example,
    OpenType file format is extended from the TrueType file format.
    """

    binding = BINDING_DEFAULT

    def __init__(self, names: IterableOf[str], *args,
                       icon: str=None, flags: int=0,
                       parents: 'IterableOf[Annotation]'=None,
                       properties: Iterable=None,
                       entities: 'IterableOf[EntityAnnotation]'=None,
                       developers: 'IterableOf[EntityAnnotation]'=None,
                       standard_entities: 'IterableOf[EntityAnnotation]'=None,
                       standard_reference: str=None,
                       standard_level: int=SPEC_STANDARD,
                       illustration: str=None,
                       extends: 'Annotation'=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.names = list(names)
        self.icon = icon
        self.flags = flags
        self.parents = list(parents or ())
        self.properties = list(properties or ())
        self.entities = list(entities or ())
        self.developers = list(developers or ())
        self.standard_entities = list(standard_entities or ())
        self.standard_reference = standard_reference
        self.standard_level = standard_level
        self.extends = extends
        self.illustration = illustration

        self.other_entities = list(self.entities)
        self.entities.extend(self.developers)
        if self.standard_entities:
            self.entities.extend(self.standard_entities)

    def __repr__(self):
        return "<%s %r at %x>" % (type(self).__name__,
                                  self.names[:1], id(self))

    @property
    def parent(self) -> 'Annotation':
        """The annotation of the first parent entity, object, or term. For example,
        TrueType font format is a type of an outline font, so its parent
        might be OUTLINE_FONT."""
        if not self.parents:
            return None
        return self.parents[0]

    def relations(self, translate: Callable=N_,
                        extended: bool=True, categories: bool=False,
                        parents: bool=True, deep: bool=True,
                        include_self: bool=True) -> 'IterableOf[tuple[str, Annotation]]':
        """Yield relations of the annotated thing with other annotated
        things, such as what it was extended from, and what it is a variant
        (child) of. Each yielded value is a tuple of an optionally translated
        string (or None) and an Annotation object describing the object of the
        relation.

        Use this to display more information about the annotation to
        the user.

        You can select which relations to include. This is recursive,
        unless deep=False is passed.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.
        """

        _ = translate

        if include_self:
            yield None, self

        if self.extends and extended:
            yield _('Extended from'), self.extends
            if deep:
                yield from self.extends.relations(translate, parents=False,
                                                  include_self=False)

        category = getattr(self, 'category', None)

        if self.parents and parents:
            for parent in self.parents:
                if parent is category and not categories:
                    continue

                yield _('A variant of'), parent
                if deep:
                    yield from parent.relations(translate, extended=False,
                                                include_self=False)

    def annotations(self) -> 'IterableOf[Annotation]':
        """Yield all annotations, including this one and its parents."""
        yield self
        for parent in self.parents:
            yield from parent.annotations()

    def ancestors(self) -> 'IterableOf[Annotation]':
        """Yield all ancestor, including the parents and their parents."""
        for parent in self.parents:
            yield parent
            yield from parent.ancestors()

    def get_name_tuples(self, prefer=PREFERRED_NAME,
                              translate: Callable=N_,
                              decorated: bool=False) -> TupleOf[str, str, Name]:
        """Yield tuples with (optionally translated) header - such as Official
        name, (optionally translated) name string, and name object.

        The returned names are ordered by the provided reference, which can
        be a flag  or a callable that receives the name object as an argument.
        One can use annotations.short or annotations.long functions here.

        If decorated=True is passed, trademark symbols may be added.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.
        """

        if not callable(prefer):
            prefer = lambda name, prefer=prefer: name.flags & prefer

        names = sorted(self.names, key=prefer, reverse=True)

        shown = 0

        for name in self.names:
            for header, flags in display_names:
                if (flags & name.flags) == flags and (shown & flags) != flags:
                    if decorated:
                        text = name.decorated_text()
                    else:
                        text = name.text
                    yield translate(header), translate(text), name
                    shown |= name.flags
                    break

    def get_name_object(self, prefer=PREFERRED_NAME) -> Name:
        """Return the best name object for the annotated thing.

        The names are ordered by the provided reference, which can
        be a flag  or a callable that receives the name object as an argument.
        One can use annotations.short or annotations.long functions here.
        """

        if not callable(prefer):
            prefer = lambda name, prefer=prefer: name.flags & prefer

        names = sorted(self.names, key=prefer, reverse=True)

        for name in names:
            return name
        return None

    def get_name(self, prefer=PREFERRED_NAME,
                       translate: Callable=N_,
                       decorated: bool=False) -> str:
        """Return the best name string for the annotated thing, optionally
        translated.

        The names are ordered by the provided reference, which can
        be a flag  or a callable that receives the name object as an argument.
        One can use annotations.short or annotations.long functions here.

        If decorated=True is passed, trademark symbols may be added.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.
        """

        name = self.get_name_object(prefer)
        if name is not None:
            if decorated:
                text = name.decorated_text()
            else:
                text = name.text
            return translate(text)
        return ''

    def get_header(self, translate: Callable=N_) -> str:
        """Return an optionally-translated string describing what the annotated
        thing is (as row/column header), such as file format, font format, etc.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.
        """
        _ = translate
        if self.flags & FORMAT_INSTANCE:
            if self.flags & DATA_FORMAT:
                return _('Font format')
            elif self.flags & FILE_FORMAT:
                return _('File format')

        if self.flags & FORMAT_SPECIFIC_CATEGORY:
            return _('Type')
        if self.flags & FORMAT_GENERIC_CATEGORY:
            return _('Category')

        # Display something useful as fallback
        return type(self).__name__

    def get_notes(self, translate: Callable=N_,
                        trademarks: bool=False) -> IteratorOf[str]:
        """Yield optionally-translated notes describes who developed
        this, who standardized it, what type of standard it is.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.

        If trademarks=True is passed, trademarks are included too.
        """
        _ = translate
        for developer in self.developers:
            yield _('Developed by %s') % (developer.get_name(), )

        if self.standard_reference:
            standard = self.standard_reference
        elif self.standard_entities:
            standard = '/'.join(entity.get_name() 
                                 for entity in self.standard_entities)
        else:
            standard = None

        if standard:
            if self.standard_level == SPEC_DRAFT:
                standard_label = _('Draft')
            elif self.standard_level == SPEC_RECOMMENDATION:
                standard_label = _('Recommendation')
            elif self.standard_level == SPEC_STANDARD:
                standard_label = _('noun|Standard')
            yield '%s; %s' % (standard_label, standard)
                                
        for other_entity in self.other_entities:
            yield _('Related entity: %s') % (other_entity.get_name(), )

        if trademarks:
            yield from self.get_trademarks(translate)

    def get_trademarks(self, translate: Callable=N_,
                             recursive: bool=True) -> IteratorOf[str]:
        """Yield trademarks strings.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.

        If recursive=True is passed, the trademarks are returned
        recursively.
        """
        _ = translate

        seen = set()

        for name in self.names:
            if not (name.flags & TRADEMARK_BIT):
                continue

            mark = name.text 
            if name.trademarked:
                mark = name.trademarked

            if mark in seen:
                continue

            seen.add(mark)

            if name.entities:
                owner = name.entities[0]
            else:
                owner = self

            if name.flags & REGISTERED_BIT:
                template = _('{name} is a registered trademark of {owner}')
            else:
                template = _('{name} is a trademark of {owner}')

            yield template.format(name=mark, 
                                  owner=owner.get_name(translate=translate, 
                                                       decorated=False))

        if recursive:
            for entity in self.entities:
                yield from entity.get_trademarks(translate, recursive)


class AnnotationWrap(object):

    """Wrap another annotation for some purpose, to provide additional
    decoration of it."""

    def __init__(self, wrapped: Annotation, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.wrapped = wrapped

    def __getattr__(self, attr):
        return getattr(self.wrapped, attr)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.wrapped)

    @property
    def wrapper_object_for(self):
        """Provide the real object to our type checker."""
        return self.wrapped


class Likely(AnnotationWrap):

    """Likely version of an annotation. This changes the binding of
    the wrapped annotation to BINDING_LIKELY. When used in a relation,
    this means that we are not sure the relation exists, only suspect
    it."""

    binding = BINDING_LIKELY


class EntityAnnotation(Annotation):
    """Anotation for entities such as organizations."""
    pass


class FormatAnnotation(Annotation):
    """Anotation for formats."""
    pass


class FileFormatAnnotation(FormatAnnotation):
    """Anotation for file formats. Respective flags are added automatically.

    This differs from a font format. The file is the container, the font is
    the data, the two can have different formats."""

    def __init__(self, *args,
                       flags: int=0,
                       font_formats: 'IterableOf[FontFormatAnnotation]'=None,
                       **kwargs):

        flags |= FILE_FORMAT

        super().__init__(*args, flags=flags, **kwargs)
        self.font_formats = list(font_formats or ())
        for font_format in self.font_formats:
            font_format.file_formats.append(self)


class FontCategoryAnnotation(FormatAnnotation):
    """Anotation for font category. Respective flags are added automatically."""

    def __init__(self, *args,
                        scalable: bool=True,
                        parametric: bool=False, flags: int=0,
                        parents: IterableOf[Annotation]=None,
                       **kwargs):

        if not (flags & (FORMAT_SPECIFIC_CATEGORY | FORMAT_GENERIC_CATEGORY)):
            if parents:
                flags |= FORMAT_SPECIFIC_CATEGORY
            else:
                flags |= FORMAT_GENERIC_CATEGORY
        flags |= DATA_FORMAT

        super().__init__(*args, flags=flags, parents=parents, **kwargs)
        self.scalable = scalable
        self.parametric = parametric


class FontFormatAnnotation(FormatAnnotation):
    """Anotation for font format. Respective flags are added automatically.

    This differs from a file format. The file is the container, the font is
    the data, the two can have different formats."""

    def __init__(self, *args,
                       category: FontCategoryAnnotation,
                       file_formats: IterableOf[FileFormatAnnotation]=None,
                       flags: int=0,
                       parents: IterableOf[Annotation]=None, **kwargs):

        flags |= DATA_FORMAT | FORMAT_INSTANCE

        parents = list(parents or ())
        parents.append(category)

        super().__init__(*args, flags=flags, parents=parents, **kwargs)
        self.category = category
        self.file_formats = list(file_formats or ())
        for file_format in self.file_formats:
            file_format.font_formats.append(self)

    def categories(self):
        return self.category.annotations()


def make_pair(*args, category: FontCategoryAnnotation,
              **kwargs) -> TupleOf[FileFormatAnnotation, FontFormatAnnotation]:
    """Create both file and font format for a thing where the two are one and
    the same. Arguments are the same as for FontFormatAnnotation."""
    file_format = FileFormatAnnotation(*args, **kwargs)
    font_format = FontFormatAnnotation(*args, category=category, **kwargs)
    file_format.font_formats.append(font_format)
    font_format.file_formats.append(file_format)
    return file_format, font_format



BITMAP_FONT = FontCategoryAnnotation([
                    Name(N_('Bitmap'), COMMON_NAME | PREFERRED_NAME),
                    Name(N_('Raster'), COMMON_NAME),
                ], icon='bitmap', illustration='bitmap-font-character.png')

OUTLINE_FONT = FontCategoryAnnotation([
                    Name(N_('Outline'), COMMON_NAME | PREFERRED_NAME),
                    Name(N_('Vector'), COMMON_NAME),
                ], icon='outline')

STROKE_FONT = FontCategoryAnnotation([
                    Name(N_('Stroke'), COMMON_NAME | PREFERRED_NAME),
                ])

CUBIC_BEZIER_FONT = FontCategoryAnnotation(
                [Name(N_('Cubic Bézier outline'), COMMON_NAME | PREFERRED_NAME)],
                parents=[OUTLINE_FONT], illustration='cubic-bezier.svgz')
QUADRATIC_BEZIER_FONT = FontCategoryAnnotation(
                [Name(N_('Quadratic Bézier outline'),
                  COMMON_NAME | PREFERRED_NAME)],
                parents=[OUTLINE_FONT], illustration='quadratic-bezier.svgz')


MICROSOFT = EntityAnnotation([
            Name('Microsoft', REGISTERED | PREFERRED_NAME),
            Name('Microsoft Corporation', REGISTERED | LEGAL_NAME,
                 trademarked='Microsoft'),
            Name('MS', PROPER_NAME | PROPRIETARY_NAME | ABBREVIATION),
            Name('Redmond', PROPER_NAME | GENERIC_NAME | UNOFFICIAL_NAME,
                 type_label=N_('Metonym')),
])

APPLE = EntityAnnotation([
            Name('Apple', REGISTERED | PREFERRED_NAME),
            Name('Apple Inc.', REGISTERED | LEGAL_NAME, trademarked='Apple'),
])


ADOBE = EntityAnnotation([
            Name('Adobe Systems', REGISTERED | PREFERRED_NAME,
                 trademarked='Adobe'),
            Name('Adobe Systems Incorporated', REGISTERED | LEGAL_NAME,
                 trademarked='Adobe'),
            Name('Adobe', REGISTERED | ABBREVIATION),
])


BITSTREAM = EntityAnnotation([
            Name('Bitstream', REGISTERED | PREFERRED_NAME),
            Name('Bitstream Inc.', REGISTERED | LEGAL_NAME,
                 trademarked='Bitstream'),
])



ISO = EntityAnnotation([
            Name('ISO', REGISTERED | ABBREVIATION | PREFERRED_NAME),
            Name(N_('International Organization for Standardization'),
                 PROPER_NAME | PROPRIETARY_NAME | FULL_NAME),
])

IEC = EntityAnnotation([
            Name('IEC', REGISTERED | ABBREVIATION | PREFERRED_NAME),
            Name(N_('International Electrotechnical Commission'),
                 PROPER_NAME | PROPRIETARY_NAME | FULL_NAME),
])


W3C = EntityAnnotation([
            Name('W3C', REGISTERED | ABBREVIATION | PREFERRED_NAME),
            Name(N_('World Wide Web Consortium'), TRADEMARK | FULL_NAME),
])



TRUETYPE_FILE = FileFormatAnnotation(
                [Name('TrueType', REGISTERED | PREFERRED_NAME,
                      entities=[APPLE]),
                 Name('TTF', PROPER_NAME | ABBREVIATION)],
                icon='ttf',
                developers=[APPLE, MICROSOFT])

OPENTYPE_FILE = FileFormatAnnotation(
                [Name('OpenType', REGISTERED | PREFERRED_NAME,
                      entities=[MICROSOFT]),
                 Name('Open Font Format', PROPER_NAME | STANDARD_NAME,
                      entities=[ISO]),
                 Name('OTF', PROPER_NAME | ABBREVIATION)],
                icon='otf',
                developers=[MICROSOFT, ADOBE],
                standard_entities=[ISO, IEC],
                standard_reference='ISO/IEC 14496-22:2015',
                extends=TRUETYPE_FILE)

TRUETYPE_FILE.parents.append(OPENTYPE_FILE)


DFONT_FILE = FileFormatAnnotation(
                [Name('Datafork TrueType', REGISTERED | PREFERRED_NAME,
                      entities=[APPLE], trademarked='TrueType'),
                 Name('dfont', PROPER_NAME | ABBREVIATION)],
                icon='ttf',
                developers=[APPLE], extends=TRUETYPE_FILE)

WOFF_FILE = FileFormatAnnotation(
                [Name('Web Open Font Format',
                      PROPER_NAME | STANDARD_NAME | PREFERRED_NAME,
                      entities=[W3C]),
                 Name('WOFF', PROPER_NAME | ABBREVIATION)],
                 icon='woff', 
                 standard_entities=[W3C],
                 standard_level=SPEC_RECOMMENDATION,
                 extends=OPENTYPE_FILE)

EMBEDDED_OPENTYPE_FILE = FileFormatAnnotation(
                [Name('Embedded OpenType', REGISTERED | PREFERRED_NAME,
                      entities=[MICROSOFT], trademarked='OpenType')],
                 icon='otf',
                 developers=[MICROSOFT],
                 extends=OPENTYPE_FILE)

POSTSCRIPT_FONT_FILE = FileFormatAnnotation(
                [Name(N_('PostScript font file'),
                      REGISTERED | PREFERRED_NAME,
                      entities=[ADOBE], trademarked='PostScript')],
                icon='ps-old',
                developers=[ADOBE])

PFA_FILE = FileFormatAnnotation(
                [Name('Printer Font ASCII',
                      PROPER_NAME | OFFICIAL_NAME | PREFERRED_NAME,
                      entities=[ADOBE]),
                 Name('PFA', PROPER_NAME | ABBREVIATION)],
                icon='ps-old',
                developers=[ADOBE],
                parents=[POSTSCRIPT_FONT_FILE])

PFB_FILE = FileFormatAnnotation(
                [Name('Printer Font Binary',
                      PROPER_NAME | OFFICIAL_NAME | PREFERRED_NAME,
                      entities=[ADOBE]),
                 Name('PFB', PROPER_NAME | ABBREVIATION)],
                icon='ps-old',
                developers=[ADOBE],
                parents=[POSTSCRIPT_FONT_FILE])


CFF_FILE = FileFormatAnnotation(
                [Name('Compact Font Format',
                      PROPER_NAME | OFFICIAL_NAME | PREFERRED_NAME,
                      entities=[ADOBE]),
                 Name('CFF', PROPER_NAME | ABBREVIATION)],
                icon='ps-old',
                developers=[ADOBE],
                parents=[POSTSCRIPT_FONT_FILE])

CEF_FILE = FileFormatAnnotation(
                [Name('Compact Embedded Font',
                      PROPER_NAME | OFFICIAL_NAME | PREFERRED_NAME,
                      entities=[ADOBE]),
                 Name('CEF', PROPER_NAME | ABBREVIATION)],
                icon='ps-old',
                developers=[ADOBE],
                parents=[POSTSCRIPT_FONT_FILE])

TRUETYPE_FONT = FontFormatAnnotation(
                [Name('TrueType', REGISTERED | PREFERRED_NAME,
                      entities=[APPLE])],
                icon='ttf', developers=[APPLE, MICROSOFT],
                category=QUADRATIC_BEZIER_FONT,
                file_formats=[TRUETYPE_FILE, OPENTYPE_FILE, WOFF_FILE,
                              DFONT_FILE, EMBEDDED_OPENTYPE_FILE])

TYPE42_FONT = FontFormatAnnotation(
                [Name('Type 42', PROPER_NAME | OFFICIAL_NAME | PREFERRED_NAME,
                      entities=[ADOBE]),
                 Name('TrueType wrapper', REGISTERED | ALT_NAME |
                                          UNOFFICIAL_NAME,
                      entities=[APPLE], trademarked='TrueType')],
                icon='ttf', developers=[APPLE, MICROSOFT],
                category=QUADRATIC_BEZIER_FONT,
                parents=[TRUETYPE_FONT],
                file_formats=[PFA_FILE, PFB_FILE, POSTSCRIPT_FONT_FILE])


POSTSCRIPT_FONT = FontFormatAnnotation(
                [Name(N_('PostScript font'),
                      PROPER_NAME | REGISTERED | PREFERRED_NAME,
                      entities=[ADOBE], trademarked='PostScript')],
                icon='ps', developers=[ADOBE],
                category=CUBIC_BEZIER_FONT,
                file_formats=[PFA_FILE, PFB_FILE, POSTSCRIPT_FONT_FILE,
                              OPENTYPE_FILE, WOFF_FILE])

TYPE1_FONT = FontFormatAnnotation(
                [Name('Type 1', PROPER_NAME | OFFICIAL_NAME | PREFERRED_NAME,
                      entities=[ADOBE])],
                icon='type1', developers=[ADOBE],
                category=CUBIC_BEZIER_FONT,
                file_formats=[PFA_FILE, PFB_FILE, POSTSCRIPT_FONT_FILE],
                parents=[POSTSCRIPT_FONT])


CFF_FONT = FontFormatAnnotation(
                [Name('Compact Font Format', PROPER_NAME | STANDARD_NAME |
                                             PREFERRED_NAME,
                      entities=[ISO]),
                 Name('CFF', PROPER_NAME | STANDARD_NAME | ABBREVIATION,
                      entities=[ISO]),
                 Name('Type 2', PROPER_NAME | OFFICIAL_NAME | ALT_NAME,
                      entities=[ADOBE])],
                icon='type2', developers=[ADOBE],
                category=CUBIC_BEZIER_FONT,
                file_formats=[OPENTYPE_FILE, WOFF_FILE,
                              DFONT_FILE, EMBEDDED_OPENTYPE_FILE,
                              CFF_FILE, CEF_FILE],
                parents=[POSTSCRIPT_FONT])


SVG_FONT = FontFormatAnnotation(
                [Name(N_('SVG color outline'),
                      PROPER_NAME | STANDARD_NAME | PREFERRED_NAME,
                      entities=[W3C]),
                 Name(N_('Scalable Vector Graphics color outline'),
                      PROPER_NAME | STANDARD_NAME | FULL_NAME,
                      entities=[W3C])],
                icon='svg', #developers=[W3C],
                standard_entities=[W3C],
                category=CUBIC_BEZIER_FONT,
                file_formats=[OPENTYPE_FILE])


BDF_FILE, BDF_FONT = make_pair(
            [Name('BDF', OFFICIAL_NAME | PREFERRED_NAME | ABBREVIATION),
             Name('Glyph Bitmap Distribution Format',
                  OFFICIAL_NAME | FULL_NAME)],
            icon='bdf', category=BITMAP_FONT)


PCF_FILE, PCF_FONT = make_pair(
            [Name('PCF', OFFICIAL_NAME | PREFERRED_NAME | ABBREVIATION),
             Name('Portable Compiled Font',
                  OFFICIAL_NAME | FULL_NAME)],
            icon='pcf', category=BITMAP_FONT)

SNF_FILE, SNF_FONT = make_pair(
            [Name('SNF', OFFICIAL_NAME | PREFERRED_NAME | ABBREVIATION),
             Name('Server Normal Format',
                  OFFICIAL_NAME | FULL_NAME)],
            icon='snf', category=BITMAP_FONT)


PFR_FILE, PFR_FONT = make_pair(
            [Name('PFR', OFFICIAL_NAME | PREFERRED_NAME | ABBREVIATION),
             Name('Portable Font Resource', OFFICIAL_NAME | FULL_NAME),
             Name('TrueDoc', REGISTERED, entities=[BITSTREAM])],
            icon='pfr', category=OUTLINE_FONT,
            developers=[BITSTREAM])


FNT_FILE, FNT_FONT = make_pair(
            [Name('Windows FNT', UNOFFICIAL_NAME)],
            icon='fnt', category=BITMAP_FONT,
            developers=[MICROSOFT])


PSF_FILE, PSF_FONT = make_pair(
            [Name('PSF', OFFICIAL_NAME | PREFERRED_NAME | ABBREVIATION),
             Name('PC Screen Font', OFFICIAL_NAME | FULL_NAME)],
            icon='psf', category=BITMAP_FONT)

SPEEDO_FILE, SPEEDO_FONT = make_pair(
            [Name('Speedo', OFFICIAL_NAME | PREFERRED_NAME | ABBREVIATION,
                  entities=[BITSTREAM])],
            icon='speedo', category=OUTLINE_FONT,
            developers=[BITSTREAM])


LEGACY_FILE, LEGACY_FONT = make_pair(
            [Name(N_('Legacy application-specific font'), UNOFFICIAL_NAME)],
            icon='legacy', category=Likely(BITMAP_FONT))



METAFONT_FILE, METAFONT_FONT = make_pair(
            [Name('Metafont', OFFICIAL_NAME | PREFERRED_NAME),
             Name('TeX font')],
            icon='metafont', category=STROKE_FONT)


